'''
Farmbot class unit tests.
'''

import sys
import json
import unittest
from unittest.mock import Mock, patch, call
import requests

from farmbot import Farmbot

MOCK_TOKEN = {
    'token': {
        'unencoded': {
            'iss': '//my.farm.bot',
            'mqtt': 'mqtt_url',
            'bot': 'device_0',
        },
        'encoded': 'encoded_token_value'
    }
}

TOKEN_REQUEST_KWARGS = {
    'headers': {'content-type': 'application/json'},
    'timeout': 0,
}

REQUEST_KWARGS_WITH_PAYLOAD = {
    'headers': {
        'authorization': 'encoded_token_value',
        'content-type': 'application/json'
    },
    'timeout': 0,
}

REQUEST_KWARGS = {
    **REQUEST_KWARGS_WITH_PAYLOAD,
    'json': None,
}


JSONDecodeError = requests.exceptions.JSONDecodeError if sys.version_info >= (
    3, 10) else json.JSONDecodeError


class TestFarmbot(unittest.TestCase):
    '''Farmbot tests'''

    def setUp(self):
        '''Set up method called before each test case'''
        self.fb = Farmbot()
        self.fb.set_token(MOCK_TOKEN)
        self.fb.set_verbosity(0)
        self.fb.state.test_env = True
        self.fb.set_timeout(0, 'all')
        self.fb.clear_cache()

    @patch('requests.post')
    def test_get_token_default_server(self, mock_post):
        '''POSITIVE TEST: function called with email, password, and default server'''
        mock_response = Mock()
        expected_token = {'token': 'abc123'}
        mock_response.json.return_value = expected_token
        mock_response.status_code = 200
        mock_response.text = 'text'
        mock_post.return_value = mock_response
        self.fb.set_token(None)
        # Call with default server
        self.fb.get_token('test_email@gmail.com', 'test_pass_123')
        mock_post.assert_called_once_with(
            url='https://my.farm.bot/api/tokens',
            **TOKEN_REQUEST_KWARGS,
            json={'user': {'email': 'test_email@gmail.com',
                           'password': 'test_pass_123'}},
        )
        self.assertEqual(self.fb.state.token, expected_token)

    @patch('requests.post')
    def test_get_token_custom_server(self, mock_post):
        '''POSITIVE TEST: function called with email, password, and custom server'''
        mock_response = Mock()
        expected_token = {'token': 'abc123'}
        mock_response.json.return_value = expected_token
        mock_response.status_code = 200
        mock_response.text = 'text'
        mock_post.return_value = mock_response
        self.fb.set_token(None)
        # Call with custom server
        self.fb.get_token('test_email@gmail.com', 'test_pass_123',
                          'https://staging.farm.bot')
        mock_post.assert_called_once_with(
            url='https://staging.farm.bot/api/tokens',
            **TOKEN_REQUEST_KWARGS,
            json={'user': {'email': 'test_email@gmail.com',
                           'password': 'test_pass_123'}},
        )
        self.assertEqual(self.fb.state.token, expected_token)

    @patch('requests.post')
    def helper_get_token_errors(self, *args, **kwargs):
        '''Test helper for get_token errors'''
        mock_post = args[0]
        status_code = kwargs['status_code']
        error_msg = kwargs['error_msg']
        mock_response = Mock()
        mock_response.status_code = status_code
        mock_post.return_value = mock_response
        self.fb.set_token(None)
        self.fb.get_token('email@gmail.com', 'test_pass_123')
        mock_post.assert_called_once_with(
            url='https://my.farm.bot/api/tokens',
            **TOKEN_REQUEST_KWARGS,
            json={'user': {'email': 'email@gmail.com',
                           'password': 'test_pass_123'}},
        )
        self.assertEqual(self.fb.state.error, error_msg)
        self.assertIsNone(self.fb.state.token)

    def test_get_token_bad_email(self):
        '''NEGATIVE TEST: function called with incorrect email'''
        self.helper_get_token_errors(
            status_code=422,
            error_msg='HTTP ERROR: Incorrect email address or password.',
        )

    def test_get_token_bad_server(self):
        '''NEGATIVE TEST: function called with incorrect server'''
        self.helper_get_token_errors(
            status_code=404,
            error_msg='HTTP ERROR: The server address does not exist.',
        )

    def test_get_token_other_error(self):
        '''get_token: other error'''
        self.helper_get_token_errors(
            status_code=500,
            error_msg='HTTP ERROR: Unexpected status code 500',
        )

    @patch('requests.post')
    def helper_get_token_exceptions(self, *args, **kwargs):
        '''Test helper for get_token exceptions'''
        mock_post = args[0]
        exception = kwargs['exception']
        error_msg = kwargs['error_msg']
        mock_post.side_effect = exception
        self.fb.set_token(None)
        self.fb.get_token('email@gmail.com', 'test_pass_123')
        mock_post.assert_called_once_with(
            url='https://my.farm.bot/api/tokens',
            **TOKEN_REQUEST_KWARGS,
            json={'user': {'email': 'email@gmail.com',
                           'password': 'test_pass_123'}},
        )
        self.assertEqual(self.fb.state.error, error_msg)
        self.assertIsNone(self.fb.state.token)

    def test_get_token_server_not_found(self):
        '''get_token: server not found'''
        self.helper_get_token_exceptions(
            exception=requests.exceptions.ConnectionError,
            error_msg='DNS ERROR: The server address does not exist.',
        )

    def test_get_token_timeout(self):
        '''get_token: timeout'''
        self.helper_get_token_exceptions(
            exception=requests.exceptions.Timeout,
            error_msg='DNS ERROR: The request timed out.',
        )

    def test_get_token_problem(self):
        '''get_token: problem'''
        self.helper_get_token_exceptions(
            exception=requests.exceptions.RequestException,
            error_msg='DNS ERROR: There was a problem with the request.',
        )

    def test_get_token_other_exception(self):
        '''get_token: other exception'''
        self.helper_get_token_exceptions(
            exception=Exception('other'),
            error_msg='DNS ERROR: An unexpected error occurred: other',
        )

    @patch('requests.request')
    def helper_api_get_error(self, *args, **kwargs):
        '''Test helper for api_get errors'''
        mock_request = args[0]
        status_code = kwargs['status_code']
        error_msg = kwargs['error_msg']
        mock_response = Mock()
        mock_response.status_code = status_code
        mock_response.reason = 'reason'
        mock_response.text = 'text'
        mock_response.json.return_value = {'error': 'error'}
        mock_request.return_value = mock_response
        response = self.fb.api_get('device')
        mock_request.assert_called_once_with(
            method='GET',
            url='https://my.farm.bot/api/device',
            **REQUEST_KWARGS,
        )
        self.assertEqual(response, error_msg)

    def test_api_get_errors(self):
        '''Test api_get errors'''
        msg_404 = 'CLIENT ERROR 404: The specified endpoint does not exist.'
        msg_404 += ' ({\n  "error": "error"\n})'
        self.helper_api_get_error(
            status_code=404,
            error_msg=msg_404
        )
        self.helper_api_get_error(
            status_code=500,
            error_msg='SERVER ERROR 500: text ({\n  "error": "error"\n})',
        )
        self.helper_api_get_error(
            status_code=600,
            error_msg='UNEXPECTED ERROR 600: text ({\n  "error": "error"\n})',
        )

    @patch('requests.request')
    def test_api_string_error_response_handling(self, mock_request):
        '''Test API string response errors'''
        mock_response = Mock()
        mock_response.status_code = 404
        mock_response.reason = 'reason'
        mock_response.text = 'error string'
        mock_response.json.side_effect = JSONDecodeError('', '', 0)
        mock_request.return_value = mock_response
        response = self.fb.api_get('device')
        mock_request.assert_called_once_with(
            method='GET',
            url='https://my.farm.bot/api/device',
            **REQUEST_KWARGS,
        )
        self.assertEqual(
            response,
            'CLIENT ERROR 404: The specified endpoint does not exist. (error string)')

    @patch('requests.request')
    def test_api_string_error_response_handling_html(self, mock_request):
        '''Test API html string response errors'''
        mock_response = Mock()
        mock_response.status_code = 404
        mock_response.reason = 'reason'
        mock_response.text = '<html><h1>error0</h1><h2>error1</h2></html>'
        mock_response.json.side_effect = JSONDecodeError('', '', 0)
        mock_request.return_value = mock_response
        response = self.fb.api_get('device')
        mock_request.assert_called_once_with(
            method='GET',
            url='https://my.farm.bot/api/device',
            **REQUEST_KWARGS,
        )
        self.assertEqual(
            response,
            'CLIENT ERROR 404: The specified endpoint does not exist. (error0 error1)')

    @patch('requests.request')
    def test_api_get_endpoint_only(self, mock_request):
        '''POSITIVE TEST: function called with endpoint only'''
        mock_response = Mock()
        expected_response = {'device': 'info'}
        mock_response.json.return_value = expected_response
        mock_response.status_code = 200
        mock_response.text = 'text'
        mock_request.return_value = mock_response
        # Call with endpoint only
        response = self.fb.api_get('device')
        mock_request.assert_called_once_with(
            method='GET',
            url='https://my.farm.bot/api/device',
            **REQUEST_KWARGS,
        )
        self.assertEqual(response, expected_response)

    @patch('requests.request')
    def test_api_get_with_payload(self, mock_request):
        '''Test api_get: with payload'''
        mock_response = Mock()
        expected_response = {'device': 'info'}
        mock_response.json.return_value = expected_response
        mock_response.status_code = 200
        mock_response.text = 'text'
        mock_request.return_value = mock_response
        # Call with endpoint only
        response = self.fb.api_get('device', payload={'key': 'value'})
        mock_request.assert_called_once_with(
            method='GET',
            url='https://my.farm.bot/api/device',
            **REQUEST_KWARGS_WITH_PAYLOAD,
            json={'key': 'value'},
        )
        self.assertEqual(response, expected_response)

    @patch('requests.request')
    def test_api_get_with_id(self, mock_request):
        '''POSITIVE TEST: function called with valid ID'''
        mock_response = Mock()
        expected_response = {'peripheral': 'info'}
        mock_response.json.return_value = expected_response
        mock_response.status_code = 200
        mock_response.text = 'text'
        mock_request.return_value = mock_response
        # Call with specific ID
        response = self.fb.api_get('peripherals', '12345')
        mock_request.assert_called_once_with(
            method='GET',
            url='https://my.farm.bot/api/peripherals/12345',
            **REQUEST_KWARGS,
        )
        self.assertEqual(response, expected_response)

    @patch('requests.request')
    def test_check_token_api_request(self, mock_request):
        '''Test check_token: API request'''
        self.fb.set_token(None)
        with self.assertRaises(ValueError) as cm:
            self.fb.api_get('points')
        self.assertEqual(cm.exception.args[0], self.fb.state.NO_TOKEN_ERROR)
        mock_request.assert_not_called()
        self.assertEqual(self.fb.state.error, self.fb.state.NO_TOKEN_ERROR)

    @patch('paho.mqtt.client.Client')
    @patch('requests.request')
    def test_check_token_broker(self, mock_request, mock_mqtt):
        '''Test check_token: broker'''
        mock_client = Mock()
        mock_mqtt.return_value = mock_client
        self.fb.set_token(None)
        with self.assertRaises(ValueError) as cm:
            self.fb.on(123)
        self.assertEqual(cm.exception.args[0], self.fb.state.NO_TOKEN_ERROR)
        with self.assertRaises(ValueError) as cm:
            self.fb.read_sensor(123)
        self.assertEqual(cm.exception.args[0], self.fb.state.NO_TOKEN_ERROR)
        with self.assertRaises(ValueError) as cm:
            self.fb.get_xyz()
        self.assertEqual(cm.exception.args[0], self.fb.state.NO_TOKEN_ERROR)
        with self.assertRaises(ValueError) as cm:
            self.fb.read_status()
        self.assertEqual(cm.exception.args[0], self.fb.state.NO_TOKEN_ERROR)
        mock_request.assert_not_called()
        mock_client.publish.assert_not_called()
        self.assertEqual(self.fb.state.error, self.fb.state.NO_TOKEN_ERROR)

    @patch('paho.mqtt.client.Client')
    def test_publish_disabled(self, mock_mqtt):
        '''Test publish disabled'''
        mock_client = Mock()
        mock_mqtt.return_value = mock_client
        self.fb.state.dry_run = True
        self.fb.on(123)
        mock_client.publish.assert_not_called()

    @patch('requests.request')
    def test_api_patch(self, mock_request):
        '''test api_patch function'''
        mock_response = Mock()
        mock_response.status_code = 200
        mock_response.text = 'text'
        mock_response.json.return_value = {'name': 'new name'}
        mock_request.return_value = mock_response
        device_info = self.fb.api_patch('device', {'name': 'new name'})
        mock_request.assert_has_calls([call(
            method='PATCH',
            url='https://my.farm.bot/api/device',
            **REQUEST_KWARGS_WITH_PAYLOAD,
            json={'name': 'new name'},
        ),
            call().json(),
        ])
        self.assertEqual(device_info, {'name': 'new name'})

    @patch('requests.request')
    def test_api_post(self, mock_request):
        '''test api_post function'''
        mock_response = Mock()
        mock_response.status_code = 200
        mock_response.text = 'text'
        mock_response.json.return_value = {'name': 'new name'}
        mock_request.return_value = mock_response
        point = self.fb.api_post('points', {'name': 'new name'})
        mock_request.assert_has_calls([call(
            method='POST',
            url='https://my.farm.bot/api/points',
            **REQUEST_KWARGS_WITH_PAYLOAD,
            json={'name': 'new name'},
        ),
            call().json(),
        ])
        self.assertEqual(point, {'name': 'new name'})

    @patch('requests.request')
    def test_api_post_no_payload(self, mock_request):
        '''test api_post: no payload'''
        mock_response = Mock()
        mock_response.status_code = 200
        mock_response.text = 'text'
        mock_response.json.return_value = {'name': 'new name'}
        mock_request.return_value = mock_response
        point = self.fb.api_post('points')
        mock_request.assert_has_calls([call(
            method='POST',
            url='https://my.farm.bot/api/points',
            **REQUEST_KWARGS,
        ),
            call().json(),
        ])
        self.assertEqual(point, {'name': 'new name'})

    @patch('requests.request')
    def test_api_delete(self, mock_request):
        '''test api_delete function'''
        mock_response = Mock()
        mock_response.status_code = 200
        mock_response.text = 'text'
        mock_response.json.return_value = {'name': 'deleted'}
        mock_request.return_value = mock_response
        result = self.fb.api_delete('points', 12345)
        mock_request.assert_called_once_with(
            method='DELETE',
            url='https://my.farm.bot/api/points/12345',
            **REQUEST_KWARGS,
        )
        self.assertEqual(result, {'name': 'deleted'})

    @patch('requests.request')
    def test_api_delete_with_payload(self, mock_request):
        '''test api_delete: with payload'''
        mock_response = Mock()
        mock_response.status_code = 200
        mock_response.text = 'text'
        mock_response.json.return_value = {'name': 'deleted'}
        mock_request.return_value = mock_response
        result = self.fb.api_delete(
            'points', 12345, payload={'key': 'value'})
        mock_request.assert_called_once_with(
            method='DELETE',
            url='https://my.farm.bot/api/points/12345',
            **REQUEST_KWARGS_WITH_PAYLOAD,
            json={'key': 'value'},
        )
        self.assertEqual(result, {'name': 'deleted'})

    @patch('requests.request')
    def test_api_delete_requests_disabled(self, mock_request):
        '''test api_delete function: requests disabled'''
        self.fb.state.dry_run = True
        result = self.fb.api_delete('points', 12345)
        mock_request.assert_not_called()
        self.assertEqual(result, {"edit_requests_disabled": True})

    @patch('requests.request')
    def helper_test_get_curve(self, *args, **kwargs):
        '''get_curve function test helper'''
        mock_request = args[0]
        mock_response = Mock()
        mock_response.json.return_value = {
            'name': 'Curve 0',
            **kwargs.get('api_data'),
        }
        mock_response.status_code = 200
        mock_response.text = 'text'
        mock_request.return_value = mock_response
        curve_info = self.fb.get_curve(12345)
        mock_request.assert_called_once_with(
            method='GET',
            url='https://my.farm.bot/api/curves/12345',
            **REQUEST_KWARGS,
        )
        self.assertEqual(curve_info['name'], 'Curve 0')
        self.assertEqual(curve_info['type'], kwargs.get('type'))
        self.assertEqual(curve_info['unit'], kwargs.get('unit'))
        self.assertEqual(curve_info.day(50), kwargs.get('value'))

    def test_get_curve(self):
        '''test get_curve function'''
        self.helper_test_get_curve(
            api_data={
                'type': 'water',
                'data': {'1': 1, '100': 100},
            },
            type='water',
            unit='mL',
            value=50,
        )
        self.helper_test_get_curve(
            api_data={
                'type': 'water',
                'data': {'100': 1, '200': 100},
            },
            type='water',
            unit='mL',
            value=1,
        )
        self.helper_test_get_curve(
            api_data={
                'type': 'water',
                'data': {'1': 1, '2': 100},
            },
            type='water',
            unit='mL',
            value=100,
        )
        self.helper_test_get_curve(
            api_data={
                'type': 'height',
                'data': {'1': 1, '50': 500, '100': 100},
            },
            type='height',
            unit='mm',
            value=500,
        )

    @patch('requests.request')
    def test_get_curve_error(self, mock_request):
        '''test get_curve function: error'''
        mock_response = Mock()
        mock_response.json.return_value = None
        mock_response.status_code = 400
        mock_response.text = 'text'
        mock_request.return_value = mock_response
        curve_info = self.fb.get_curve(12345)
        mock_request.assert_called_once_with(
            method='GET',
            url='https://my.farm.bot/api/curves/12345',
            **REQUEST_KWARGS,
        )
        self.assertIsNone(curve_info)

    @patch('requests.request')
    def test_safe_z(self, mock_request):
        '''test safe_z function'''
        mock_response = Mock()
        mock_response.json.return_value = {'safe_height': 100}
        mock_response.status_code = 200
        mock_response.text = 'text'
        mock_request.return_value = mock_response
        safe_height = self.fb.safe_z()
        mock_request.assert_called_once_with(
            method='GET',
            url='https://my.farm.bot/api/fbos_config',
            **REQUEST_KWARGS,
        )
        self.assertEqual(safe_height, 100)

    @patch('requests.request')
    def test_garden_size(self, mock_request):
        '''test garden_size function'''
        mock_response = Mock()
        mock_response.json.return_value = {
            'movement_axis_nr_steps_x': 1000,
            'movement_axis_nr_steps_y': 2000,
            'movement_axis_nr_steps_z': 40000,
            'movement_step_per_mm_x': 5,
            'movement_step_per_mm_y': 5,
            'movement_step_per_mm_z': 25,
        }
        mock_response.status_code = 200
        mock_response.text = 'text'
        mock_request.return_value = mock_response
        garden_size = self.fb.garden_size()
        mock_request.assert_called_once_with(
            method='GET',
            url='https://my.farm.bot/api/firmware_config',
            **REQUEST_KWARGS,
        )
        self.assertEqual(garden_size, {'x': 200, 'y': 400, 'z': 1600})

    @patch('requests.request')
    def test_log(self, mock_request):
        '''test log function'''
        mock_response = Mock()
        mock_response.status_code = 200
        mock_response.text = 'text'
        mock_response.json.return_value = {'message': 'test message'}
        mock_request.return_value = mock_response
        self.fb.log('test message', 'info', ['toast'])
        mock_request.assert_called_once_with(
            method='POST',
            url='https://my.farm.bot/api/logs',
            **REQUEST_KWARGS_WITH_PAYLOAD,
            json={
                'message': 'test message',
                'type': 'info',
                'channels': ['toast'],
            },
        )

    @patch('paho.mqtt.client.Client')
    def test_connect_broker(self, mock_mqtt):
        '''Test test_connect_broker command'''
        mock_client = Mock()
        mock_mqtt.return_value = mock_client
        self.fb.connect_broker()
        mock_client.username_pw_set.assert_called_once_with(
            username='device_0',
            password='encoded_token_value')
        mock_client.connect.assert_called_once_with(
            'mqtt_url',
            port=1883,
            keepalive=60)
        mock_client.loop_start.assert_called()

    def test_disconnect_broker(self):
        '''Test disconnect_broker command'''
        mock_client = Mock()
        self.fb.broker.client = mock_client
        self.fb.disconnect_broker()
        mock_client.loop_stop.assert_called_once()
        mock_client.disconnect.assert_called_once()

    @patch('paho.mqtt.client.Client')
    def test_listen(self, mock_mqtt):
        '''Test listen command'''
        mock_client = Mock()
        mock_mqtt.return_value = mock_client
        self.fb.listen()

        class MockMessage:
            '''Mock message class'''
            topic = 'bot/device_0/topic'
            payload = '{"message": "test message"}'
        mock_client.on_message('', '', MockMessage())
        mock_client.username_pw_set.assert_called_once_with(
            username='device_0',
            password='encoded_token_value')
        mock_client.connect.assert_called_once_with(
            'mqtt_url',
            port=1883,
            keepalive=60)
        mock_client.subscribe.assert_called_once_with('bot/device_0/#')
        mock_client.loop_start.assert_called()
        mock_client.loop_stop.assert_called()
        self.assertEqual(self.fb.state.last_messages['topic'], [{
            'topic': 'bot/device_0/topic',
            'content': {'message': 'test message'},
        }])

    @patch('paho.mqtt.client.Client')
    def test_listen_diff_only(self, mock_mqtt):
        '''Test listen command: diff_only'''
        mock_client = Mock()
        mock_mqtt.return_value = mock_client
        self.fb.listen(message_options={'diff_only': True})

        class MockMessageFirst:
            '''Mock message class'''
            topic = 'bot/device_0/topic'
            payload = '{"message": "test message", "i": 0}'
        mock_client.on_message('', '', MockMessageFirst())

        class MockMessageSecond:
            '''Mock message class'''
            topic = 'bot/device_0/topic'
            payload = '{"message": "test message", "i": 1}'
        mock_client.on_message('', '', MockMessageSecond())
        mock_client.username_pw_set.assert_called_once_with(
            username='device_0',
            password='encoded_token_value')
        mock_client.connect.assert_called_once_with(
            'mqtt_url',
            port=1883,
            keepalive=60)
        mock_client.subscribe.assert_called_once_with('bot/device_0/#')
        mock_client.loop_start.assert_called()
        mock_client.loop_stop.assert_called()
        self.assertEqual(self.fb.state.last_messages['topic'], [
            {'topic': 'bot/device_0/topic',
             'content': {'message': 'test message', 'i': 0}},
            {'topic': 'bot/device_0/topic',
             'content': {'message': 'test message', 'i': 1}},
        ])
        self.assertEqual(self.fb.state.last_messages['topic_diffs'], [
            {'i': 1},
        ])

    @patch('paho.mqtt.client.Client')
    def test_listen_with_filters(self, mock_mqtt):
        '''Test listen command with filters'''
        mock_client = Mock()
        mock_mqtt.return_value = mock_client
        self.fb.listen(message_options={'filters': {
            'topic': 'sync/Point',
            'content': {'body.pointerType': 'Plant'}}})

        class MockMessageMiss:
            '''Mock message class'''
            topic = 'bot/device_0/topic'
            payload = '{"message": "test message"}'
        mock_client.on_message('', '', MockMessageMiss())

        class MockMessageAlsoMiss:
            '''Mock message class'''
            topic = 'bot/device_0/sync/Point/123'
            payload = json.dumps({'body': {'pointerType': 'Weed'}})
        mock_client.on_message('', '', MockMessageAlsoMiss())

        class MockMessageMatch:
            '''Mock message class'''
            topic = 'bot/device_0/sync/Point/1234'
            payload = json.dumps({'body': {'pointerType': 'Plant'}})
        mock_client.on_message('', '', MockMessageMatch())
        mock_client.username_pw_set.assert_called_once_with(
            username='device_0',
            password='encoded_token_value')
        mock_client.connect.assert_called_once_with(
            'mqtt_url',
            port=1883,
            keepalive=60)
        mock_client.subscribe.assert_called_once_with('bot/device_0/#')
        mock_client.loop_start.assert_called()
        mock_client.loop_stop.assert_called()
        self.assertEqual(self.fb.state.last_messages['sync'], [{
            'topic': 'bot/device_0/sync/Point/1234',
            'content': {'body': {'pointerType': 'Plant'}},
        }])

    @patch('math.inf', 0.1)
    @patch('paho.mqtt.client.Client')
    def test_listen_for_status_changes(self, mock_mqtt):
        '''Test listen_for_status_changes command'''
        self.maxDiff = None
        i = 0

        mock_client = Mock()
        mock_mqtt.return_value = mock_client

        class MockMessage:
            '''Mock message class'''

            def __init__(self):
                self.topic = 'bot/device_0/status'
                payload = {
                    'location_data': {
                        'position': {
                            'x': i,
                            'y': i + 10,
                            'z': 100,
                        }}}
                if i == 2:
                    payload['location_data']['position']['extra'] = {'idx': 2}
                if i == 3:
                    payload['location_data']['position']['extra'] = {'idx': 3}
                self.payload = json.dumps(payload)

        def patched_sleep(_seconds):
            '''Patched sleep function'''
            nonlocal i
            mock_message = MockMessage()
            mock_client.on_message('', '', mock_message)
            i += 1

        with patch('time.sleep', new=patched_sleep):
            self.fb.listen_for_status_changes(
                stop_count=5,
                path='location_data.position')

        self.assertEqual(self.fb.state.last_messages['status'], [
            {
                'topic': 'bot/device_0/status',
                'content': {'location_data': {'position': {'x': 0, 'y': 10, 'z': 100}}}},
            {
                'topic': 'bot/device_0/status',
                'content': {'location_data': {'position': {'x': 1, 'y': 11, 'z': 100}}}},
            {
                'topic': 'bot/device_0/status',
                'content': {'location_data': {'position': {
                    'extra': {'idx': 2}, 'x': 2, 'y': 12, 'z': 100}}}},
            {
                'topic': 'bot/device_0/status',
                'content': {'location_data': {'position': {
                    'extra': {'idx': 3}, 'x': 3, 'y': 13, 'z': 100}}}},
            {
                'topic': 'bot/device_0/status',
                'content': {'location_data': {'position': {'x': 4, 'y': 14, 'z': 100}}}},
        ])
        self.assertEqual(self.fb.state.last_messages['status_diffs'], [
            {'x': 1, 'y': 11},
            {'extra': {'idx': 2}, 'x': 2, 'y': 12},
            {'extra': {'idx': 3}, 'x': 3, 'y': 13},
            {'x': 4, 'y': 14},
        ])
        self.assertEqual(self.fb.state.last_messages['status_excerpt'], [
            {'x': 0, 'y': 10, 'z': 100},
            {'x': 1, 'y': 11, 'z': 100},
            {'extra': {'idx': 2}, 'x': 2, 'y': 12, 'z': 100},
            {'extra': {'idx': 3}, 'x': 3, 'y': 13, 'z': 100},
            {'x': 4, 'y': 14, 'z': 100},
        ])

    @patch('math.inf', 0.1)
    @patch('paho.mqtt.client.Client')
    def test_listen_for_status_changes_single_value(self, mock_mqtt):
        '''Test listen_for_status_changes command: single value'''
        self.maxDiff = None
        i = 0

        mock_client = Mock()
        mock_mqtt.return_value = mock_client

        class MockMessage:
            '''Mock message class'''

            def __init__(self):
                self.topic = 'bot/device_0/status'
                payload = {'location_data': {'position': {'x': i}}}
                self.payload = json.dumps(payload)

        def patched_sleep(_seconds):
            '''Patched sleep function'''
            nonlocal i
            mock_message = MockMessage()
            mock_client.on_message('', '', mock_message)
            i += 1

        with patch('time.sleep', new=patched_sleep):
            self.fb.listen_for_status_changes(
                stop_count=5,
                path='location_data.position.x')

        self.assertEqual(self.fb.state.last_messages['status_diffs'],
                         [1, 2, 3, 4])
        self.assertEqual(self.fb.state.last_messages['status_excerpt'],
                         [0, 1, 2, 3, 4])

    @patch('paho.mqtt.client.Client')
    def test_listen_clear_last(self, mock_mqtt):
        '''Test listen command: clear last message'''
        mock_client = Mock()
        mock_mqtt.return_value = mock_client
        self.fb.state.last_messages = [
            {'#': {'topic': '', 'content': "message"}},
        ]
        self.fb.state.test_env = False
        self.fb.listen()
        self.assertEqual(len(self.fb.state.last_messages['#']), 0)

    @patch('paho.mqtt.client.Client')
    def test_listen_interrupt(self, mock_mqtt):
        '''Test listen command: interrupt'''
        mock_client = Mock()
        mock_mqtt.return_value = mock_client
        self.fb.state.test_env = False

        def patched_sleep(_seconds):
            '''Patched sleep function'''
            raise KeyboardInterrupt
        with patch('time.sleep', new=patched_sleep):
            self.fb.listen(stop_count=100)
        mock_client.loop_stop.assert_called_once()

    @patch('paho.mqtt.client.Client')
    def test_publish_apply_label(self, mock_mqtt):
        '''Test publish command: set uuid'''
        mock_client = Mock()
        mock_mqtt.return_value = mock_client
        self.fb.state.test_env = False
        self.fb.publish({'kind': 'sync', 'args': {}})
        label = self.fb.state.last_published.get('args', {}).get('label')
        self.assertNotIn(label, ['test', '', None])

    @patch('requests.request')
    @patch('paho.mqtt.client.Client')
    def send_command_test_helper(self, *args, **kwargs):
        '''Helper for testing command execution'''
        execute_command = args[0]
        mock_mqtt = args[1]
        mock_request = args[2]
        expected_command = kwargs.get('expected_command')
        extra_rpc_args = kwargs.get('extra_rpc_args')
        mock_api_response = kwargs.get('mock_api_response')
        error = kwargs.get('error')
        mock_client = Mock()
        mock_mqtt.return_value = mock_client
        mock_response = Mock()
        mock_response.json.return_value = mock_api_response
        mock_response.status_code = 200
        mock_response.text = 'text'
        mock_request.return_value = mock_response
        self.fb.state.last_messages['from_device'] = [{
            'topic': '',
            'content': {
                'kind': 'rpc_error' if error else 'rpc_ok',
                'args': {'label': 'test'},
            },
        }]
        execute_command()
        if expected_command is None:
            mock_client.publish.assert_not_called()
            return
        expected_payload = {
            'kind': 'rpc_request',
            'args': {'label': 'test', **extra_rpc_args},
            'body': [expected_command],
        }
        mock_client.username_pw_set.assert_called_once_with(
            username='device_0',
            password='encoded_token_value')
        mock_client.connect.assert_called_once_with(
            'mqtt_url',
            port=1883,
            keepalive=60)
        mock_client.loop_start.assert_called()
        mock_client.publish.assert_called_once_with(
            'bot/device_0/from_clients',
            payload=json.dumps(expected_payload))
        if not error:
            self.assertNotEqual(
                self.fb.state.error,
                'RPC error response received.')

    def test_message(self):
        '''Test message command'''
        def exec_command():
            self.fb.send_message('test message', 'info')
        self.send_command_test_helper(
            exec_command,
            expected_command={
                'kind': 'send_message',
                'args': {'message': 'test message', 'message_type': 'info'},
                'body': [],
            },
            extra_rpc_args={},
            mock_api_response={})

    def test_debug(self):
        '''Test debug command'''
        def exec_command():
            self.fb.debug('test message')
        self.send_command_test_helper(
            exec_command,
            expected_command={
                'kind': 'send_message',
                'args': {'message': 'test message', 'message_type': 'debug'},
                'body': [],
            },
            extra_rpc_args={},
            mock_api_response={})

    def test_toast(self):
        '''Test toast command'''
        def exec_command():
            self.fb.toast('test message')
        self.send_command_test_helper(
            exec_command,
            expected_command={
                'kind': 'send_message',
                'args': {'message': 'test message', 'message_type': 'info'},
                'body': [{'kind': 'channel', 'args': {'channel_name': 'toast'}}],
            },
            extra_rpc_args={},
            mock_api_response={})

    def test_invalid_message_type(self):
        '''Test message_type validation'''
        def exec_command():
            with self.assertRaises(ValueError) as cm:
                self.fb.send_message('test', message_type='nope')
            msg = 'Invalid message type: `nope` not in '
            msg += "['assertion', 'busy', 'debug', 'error', 'fun', 'info', 'success', 'warn']"
            self.assertEqual(cm.exception.args[0], msg)
        self.send_command_test_helper(
            exec_command,
            expected_command=None,
            extra_rpc_args={},
            mock_api_response={})

    def test_invalid_message_channel(self):
        '''Test message channel validation'''
        def exec_command():
            with self.assertRaises(ValueError) as cm:
                self.fb.send_message('test', channels=['nope'])
            self.assertEqual(
                cm.exception.args[0],
                "Invalid channel: nope not in ['ticker', 'toast', 'email', 'espeak']")
        self.send_command_test_helper(
            exec_command,
            expected_command=None,
            extra_rpc_args={},
            mock_api_response={})

    def test_read_status(self):
        '''Test read_status command'''
        def exec_command():
            self.fb.state.last_messages['status'] = [{
                'topic': '',
                'content': {'location_data': {'position': {'x': 100}}},
            }]
            result = self.fb.read_status()
            self.assertEqual(
                result,
                {'location_data': {'position': {'x': 100}}})
        self.send_command_test_helper(
            exec_command,
            expected_command={
                'kind': 'read_status',
                'args': {},
            },
            extra_rpc_args={},
            mock_api_response={})

    def test_read_status_path(self):
        '''Test read_status command: specific path'''
        def exec_command():
            self.fb.state.last_messages['status'] = [{
                'topic': '',
                'content': {
                    'location_data': {'position': {'x': 100}},
                },
            }]
            result = self.fb.read_status('location_data.position.x')
            self.assertEqual(result, 100)
        self.send_command_test_helper(
            exec_command,
            expected_command={
                'kind': 'read_status',
                'args': {},
            },
            extra_rpc_args={},
            mock_api_response={})

    def test_read_pin(self):
        '''Test read_pin command'''
        def exec_command():
            self.fb.read_pin(13)
        self.send_command_test_helper(
            exec_command,
            expected_command={
                'kind': 'read_pin',
                'args': {
                    'pin_number': 13,
                    'label': '---',
                    'pin_mode': 0,
                },
            },
            extra_rpc_args={},
            mock_api_response={})

    def test_read_sensor(self):
        '''Test read_sensor command'''
        def exec_command():
            self.fb.read_sensor('Tool Verification')
        self.send_command_test_helper(
            exec_command,
            expected_command={
                'kind': 'read_pin',
                'args': {
                    'pin_mode': 0,
                    'label': '---',
                    'pin_number': {
                        'kind': 'named_pin',
                        'args': {'pin_type': 'Sensor', 'pin_id': 123},
                    },
                },
            },
            extra_rpc_args={},
            mock_api_response=[{'id': 123, 'label': 'Tool Verification', 'mode': 0}])

    def test_read_sensor_not_found(self):
        '''Test read_sensor command: sensor not found'''
        def exec_command():
            self.fb.read_sensor('Temperature')
        self.send_command_test_helper(
            exec_command,
            expected_command=None,
            extra_rpc_args={},
            mock_api_response=[{'label': 'Tool Verification'}])
        self.assertEqual(
            self.fb.state.error,
            "ERROR: 'Temperature' not in sensors: ['Tool Verification'].")

    def test_assertion(self):
        '''Test assertion command'''
        def exec_command():
            self.fb.assertion('return true', 'abort')
        self.send_command_test_helper(
            exec_command,
            expected_command={
                'kind': 'assertion',
                'args': {
                    'assertion_type': 'abort',
                    'lua': 'return true',
                    '_then': {'kind': 'nothing', 'args': {}},
                }
            },
            extra_rpc_args={},
            mock_api_response={})

    def test_assertion_with_recovery_sequence(self):
        '''Test assertion command with recovery sequence'''
        def exec_command():
            self.fb.assertion('return true', 'abort', 'Recovery Sequence')
        self.send_command_test_helper(
            exec_command,
            expected_command={
                'kind': 'assertion',
                'args': {
                    'assertion_type': 'abort',
                    'lua': 'return true',
                    '_then': {'kind': 'execute', 'args': {'sequence_id': 123}},
                }
            },
            extra_rpc_args={},
            mock_api_response=[{'id': 123, 'name': 'Recovery Sequence'}])

    def test_assertion_recovery_sequence_not_found(self):
        '''Test assertion command: recovery sequence not found'''
        def exec_command():
            self.fb.assertion('return true', 'abort', 'Recovery Sequence')
        self.send_command_test_helper(
            exec_command,
            expected_command=None,
            extra_rpc_args={},
            mock_api_response=[])
        self.assertEqual(
            self.fb.state.error,
            "ERROR: 'Recovery Sequence' not in sequences: [].")

    def test_assertion_invalid_assertion_type(self):
        '''Test assertion command: invalid assertion type'''
        def exec_command():
            with self.assertRaises(ValueError) as cm:
                self.fb.assertion('return true', 'nope')
            msg = 'Invalid assertion_type: nope not in '
            msg += "['abort', 'recover', 'abort_recover', 'continue']"
            self.assertEqual(cm.exception.args[0], msg)
        self.send_command_test_helper(
            exec_command,
            expected_command=None,
            extra_rpc_args={},
            mock_api_response={})

    def test_wait(self):
        '''Test wait command'''
        def exec_command():
            self.fb.wait(123)
        self.send_command_test_helper(
            exec_command,
            expected_command={
                'kind': 'wait',
                'args': {'milliseconds': 123},
            },
            extra_rpc_args={},
            mock_api_response={})

    def test_unlock(self):
        '''Test unlock command'''
        def exec_command():
            self.fb.unlock()
        self.send_command_test_helper(
            exec_command,
            expected_command={
                'kind': 'emergency_unlock',
                'args': {},
            },
            extra_rpc_args={'priority': 9000},
            mock_api_response={})

    def test_e_stop(self):
        '''Test e_stop command'''
        def exec_command():
            self.fb.e_stop()
        self.send_command_test_helper(
            exec_command,
            expected_command={
                'kind': 'emergency_lock',
                'args': {},
            },
            extra_rpc_args={'priority': 9000},
            mock_api_response={})

    def test_find_home(self):
        '''Test find_home command'''
        def exec_command():
            self.fb.find_home()
        self.send_command_test_helper(
            exec_command,
            expected_command={
                'kind': 'find_home',
                'args': {'axis': 'all', 'speed': 100},
            },
            extra_rpc_args={},
            mock_api_response={})

    def test_find_home_speed_error(self):
        '''Test find_home command: speed error'''
        def exec_command():
            self.fb.find_home('all', 0)
        self.send_command_test_helper(
            exec_command,
            expected_command=None,
            extra_rpc_args={},
            mock_api_response={})
        self.assertEqual(
            self.fb.state.error,
            'ERROR: Speed constrained to 1-100.')

    def test_find_home_invalid_axis(self):
        '''Test find_home command: invalid axis'''
        def exec_command():
            with self.assertRaises(ValueError) as cm:
                self.fb.find_home('nope')
            self.assertEqual(
                cm.exception.args[0],
                "Invalid axis: nope not in ['x', 'y', 'z', 'all']")
        self.send_command_test_helper(
            exec_command,
            expected_command=None,
            extra_rpc_args={},
            mock_api_response={})

    def test_set_home(self):
        '''Test set_home command'''
        def exec_command():
            self.fb.set_home()
        self.send_command_test_helper(
            exec_command,
            expected_command={
                'kind': 'zero',
                'args': {'axis': 'all'},
            },
            extra_rpc_args={},
            mock_api_response={})

    def test_toggle_peripheral(self):
        '''Test toggle_peripheral command'''
        def exec_command():
            self.fb.toggle_peripheral('New Peripheral')
        self.send_command_test_helper(
            exec_command,
            expected_command={
                'kind': 'toggle_pin',
                'args': {
                    'pin_number': {
                        'kind': 'named_pin',
                        'args': {'pin_type': 'Peripheral', 'pin_id': 123},
                    },
                },
            },
            extra_rpc_args={},
            mock_api_response=[{'label': 'New Peripheral', 'id': 123}])

    def test_toggle_peripheral_not_found(self):
        '''Test toggle_peripheral command: peripheral not found'''
        def exec_command():
            self.fb.toggle_peripheral('New Peripheral')
        self.send_command_test_helper(
            exec_command,
            expected_command=None,
            extra_rpc_args={},
            mock_api_response=[])
        self.assertEqual(
            self.fb.state.error,
            'ERROR: \'New Peripheral\' not in peripherals: [].')

    @patch('requests.request')
    @patch('paho.mqtt.client.Client')
    def test_toggle_peripheral_use_cache(self, mock_mqtt, mock_request):
        '''Test toggle_peripheral command: use cache'''
        mock_client = Mock()
        mock_mqtt.return_value = mock_client
        mock_response = Mock()
        mock_response.json.return_value = [
            {'label': 'Peripheral 4', 'id': 123},
            {'label': 'Peripheral 5', 'id': 456}
        ]
        mock_response.status_code = 200
        mock_response.text = 'text'
        mock_request.return_value = mock_response
        # save cache
        self.fb.toggle_peripheral('Peripheral 4')
        mock_request.assert_called()
        mock_client.publish.assert_called()
        mock_request.reset_mock()
        mock_client.reset_mock()
        # use cache
        self.fb.toggle_peripheral('Peripheral 5')
        mock_request.assert_not_called()
        mock_client.publish.assert_called()
        mock_request.reset_mock()
        mock_client.reset_mock()
        # clear cache
        self.fb.toggle_peripheral('Peripheral 6')
        mock_request.assert_not_called()
        mock_client.publish.assert_not_called()
        mock_request.reset_mock()
        mock_client.reset_mock()
        # save cache
        self.fb.toggle_peripheral('Peripheral 4')
        mock_request.assert_called()
        mock_client.publish.assert_called()
        mock_request.reset_mock()
        mock_client.reset_mock()

    def test_on_digital(self):
        '''Test on command: digital'''
        def exec_command():
            self.fb.on(13)
        self.send_command_test_helper(
            exec_command,
            expected_command={
                'kind': 'write_pin',
                'args': {
                    'pin_value': 1,
                    'pin_mode': 0,
                    'pin_number': 13,
                },
            },
            extra_rpc_args={},
            mock_api_response={})

    def test_off(self):
        '''Test off command'''
        def exec_command():
            self.fb.off(13)
        self.send_command_test_helper(
            exec_command,
            expected_command={
                'kind': 'write_pin',
                'args': {
                    'pin_value': 0,
                    'pin_mode': 0,
                    'pin_number': 13,
                },
            },
            extra_rpc_args={},
            mock_api_response={})

    def test_move(self):
        '''Test move command'''
        def exec_command():
            self.fb.move()
        self.send_command_test_helper(
            exec_command,
            expected_command={
                'kind': 'move',
                'args': {},
                'body': [],
            },
            extra_rpc_args={},
            mock_api_response={})

    def test_move_extras(self):
        '''Test move command with extras'''
        def exec_command():
            self.fb.move(1, 2, 3, safe_z=True, speed=50)
        self.send_command_test_helper(
            exec_command,
            expected_command={
                'kind': 'move',
                'args': {},
                'body': [
                    {'kind': 'axis_overwrite', 'args': {
                        'axis': 'x',
                        'axis_operand': {'kind': 'numeric', 'args': {'number': 1}}}},
                    {'kind': 'axis_overwrite', 'args': {
                        'axis': 'y',
                        'axis_operand': {'kind': 'numeric', 'args': {'number': 2}}}},
                    {'kind': 'axis_overwrite', 'args': {
                        'axis': 'z',
                        'axis_operand': {'kind': 'numeric', 'args': {'number': 3}}}},
                    {'kind': 'speed_overwrite', 'args': {
                        'axis': 'x',
                        'speed_setting': {'kind': 'numeric', 'args': {'number': 50}}}},
                    {'kind': 'speed_overwrite', 'args': {
                        'axis': 'y',
                        'speed_setting': {'kind': 'numeric', 'args': {'number': 50}}}},
                    {'kind': 'speed_overwrite', 'args': {
                        'axis': 'z',
                        'speed_setting': {'kind': 'numeric', 'args': {'number': 50}}}},
                    {'kind': 'safe_z', 'args': {}},
                ],
            },
            extra_rpc_args={},
            mock_api_response={})

    def test_reboot(self):
        '''Test reboot command'''
        def exec_command():
            self.fb.reboot()
        self.send_command_test_helper(
            exec_command,
            expected_command={
                'kind': 'reboot',
                'args': {'package': 'farmbot_os'},
            },
            extra_rpc_args={},
            mock_api_response={})

    def test_shutdown(self):
        '''Test shutdown command'''
        def exec_command():
            self.fb.shutdown()
        self.send_command_test_helper(
            exec_command,
            expected_command={
                'kind': 'power_off',
                'args': {},
            },
            extra_rpc_args={},
            mock_api_response={})

    def test_find_axis_length(self):
        '''Test find_axis_length command'''
        def exec_command():
            self.fb.find_axis_length()
        self.send_command_test_helper(
            exec_command,
            expected_command={
                'kind': 'calibrate',
                'args': {'axis': 'all'},
            },
            extra_rpc_args={},
            mock_api_response={})

    def test_write_pin(self):
        '''Test write_pin command'''
        def exec_command():
            self.fb.write_pin(13, 1, 'analog')
        self.send_command_test_helper(
            exec_command,
            expected_command={
                'kind': 'write_pin',
                'args': {
                    'pin_number': 13,
                    'pin_value': 1,
                    'pin_mode': 1,
                },
            },
            extra_rpc_args={},
            mock_api_response={})

    def test_write_pin_invalid_mode(self):
        '''Test write_pin command: invalid mode'''
        def exec_command():
            with self.assertRaises(ValueError) as cm:
                self.fb.write_pin(13, 1, 1)
            self.assertEqual(
                cm.exception.args[0],
                "Invalid mode: 1 not in ['digital', 'analog']")
        self.send_command_test_helper(
            exec_command,
            expected_command=None,
            extra_rpc_args={},
            mock_api_response={})

    def test_control_peripheral(self):
        '''Test control_peripheral command'''
        def exec_command():
            self.fb.control_peripheral('New Peripheral', 1)
        self.send_command_test_helper(
            exec_command,
            expected_command={
                'kind': 'write_pin',
                'args': {
                    'pin_value': 1,
                    'pin_mode': 0,
                    'pin_number': {
                        'kind': 'named_pin',
                        'args': {'pin_type': 'Peripheral', 'pin_id': 123},
                    },
                },
            },
            extra_rpc_args={},
            mock_api_response=[{'label': 'New Peripheral', 'mode': 0, 'id': 123}])

    def test_control_peripheral_analog(self):
        '''Test control_peripheral command: analog'''
        def exec_command():
            self.fb.control_peripheral('New Peripheral', 1, 'analog')
        self.send_command_test_helper(
            exec_command,
            expected_command={
                'kind': 'write_pin',
                'args': {
                    'pin_value': 1,
                    'pin_mode': 1,
                    'pin_number': {
                        'kind': 'named_pin',
                        'args': {'pin_type': 'Peripheral', 'pin_id': 123},
                    },
                },
            },
            extra_rpc_args={},
            mock_api_response=[{'label': 'New Peripheral', 'mode': 0, 'id': 123}])

    def test_control_peripheral_not_found(self):
        '''Test control_peripheral command: peripheral not found'''
        def exec_command():
            self.fb.control_peripheral('New Peripheral', 1)
        self.send_command_test_helper(
            exec_command,
            expected_command=None,
            extra_rpc_args={},
            mock_api_response=[{'label': 'Pump'}, {'label': 'Lights'}])
        self.assertEqual(
            self.fb.state.error,
            "ERROR: 'New Peripheral' not in peripherals: ['Pump', 'Lights'].")

    def test_measure_soil_height(self):
        '''Test measure_soil_height command'''
        def exec_command():
            self.fb.measure_soil_height()
        self.send_command_test_helper(
            exec_command,
            expected_command={
                'kind': 'execute_script',
                'args': {'label': 'Measure Soil Height'},
            },
            extra_rpc_args={},
            mock_api_response={})

    def test_detect_weeds(self):
        '''Test detect_weeds command'''
        def exec_command():
            self.fb.detect_weeds()
        self.send_command_test_helper(
            exec_command,
            expected_command={
                'kind': 'execute_script',
                'args': {'label': 'plant-detection'},
            },
            extra_rpc_args={},
            mock_api_response={})

    def test_calibrate_camera(self):
        '''Test calibrate_camera command'''
        def exec_command():
            self.fb.calibrate_camera()
        self.send_command_test_helper(
            exec_command,
            expected_command={
                'kind': 'execute_script',
                'args': {'label': 'camera-calibration'},
            },
            extra_rpc_args={},
            mock_api_response={})

    def test_sequence(self):
        '''Test sequence command'''
        def exec_command():
            self.fb.sequence('My Sequence')
        self.send_command_test_helper(
            exec_command,
            expected_command={
                'kind': 'execute',
                'args': {'sequence_id': 123},
            },
            extra_rpc_args={},
            mock_api_response=[{'name': 'My Sequence', 'id': 123}])

    def test_sequence_not_found(self):
        '''Test sequence command: sequence not found'''
        def exec_command():
            self.fb.sequence('My Sequence')
        self.send_command_test_helper(
            exec_command,
            expected_command=None,
            extra_rpc_args={},
            mock_api_response=[{'name': 'Water'}])
        self.assertEqual(
            self.fb.state.error,
            "ERROR: 'My Sequence' not in sequences: ['Water'].")

    def test_take_photo(self):
        '''Test take_photo command'''
        def exec_command():
            self.fb.take_photo()
        self.send_command_test_helper(
            exec_command,
            expected_command={
                'kind': 'take_photo',
                'args': {},
            },
            extra_rpc_args={},
            mock_api_response={})

    def test_control_servo(self):
        '''Test control_servo command'''
        def exec_command():
            self.fb.control_servo(4, 100)
        self.send_command_test_helper(
            exec_command,
            expected_command={
                'kind': 'set_servo_angle',
                'args': {
                    'pin_number': 4,
                    'pin_value': 100,
                },
            },
            extra_rpc_args={},
            mock_api_response={'mode': 0})

    def test_control_servo_error(self):
        '''Test control_servo command: error'''
        def exec_command():
            self.fb.control_servo(4, 200)
        self.send_command_test_helper(
            exec_command,
            expected_command=None,
            extra_rpc_args={},
            mock_api_response={'mode': 0})

    def test_get_xyz(self):
        '''Test get_xyz command'''
        def exec_command():
            self.fb.state.last_messages['status'] = [{
                'topic': '',
                'content': {
                    'location_data': {'position': {'x': 1, 'y': 2, 'z': 3}},
                },
            }]
            position = self.fb.get_xyz()
            self.assertEqual(position, {'x': 1, 'y': 2, 'z': 3})
        self.send_command_test_helper(
            exec_command,
            expected_command={
                'kind': 'read_status',
                'args': {},
            },
            extra_rpc_args={},
            mock_api_response={})

    def test_get_xyz_no_status(self):
        '''Test get_xyz command: no status'''
        def exec_command():
            self.fb.state.last_messages['status'] = []
            position = self.fb.get_xyz()
            self.assertIsNone(position)
        self.send_command_test_helper(
            exec_command,
            expected_command={
                'kind': 'read_status',
                'args': {},
            },
            extra_rpc_args={},
            mock_api_response={})

    def test_check_position(self):
        '''Test check_position command: at position'''
        def exec_command():
            self.fb.state.last_messages['status'] = [{
                'topic': '',
                'content': {
                    'location_data': {'position': {'x': 1, 'y': 2, 'z': 3}},
                },
            }]
            at_position = self.fb.check_position({'x': 1, 'y': 2, 'z': 3}, 0)
            self.assertTrue(at_position)
        self.send_command_test_helper(
            exec_command,
            expected_command={
                'kind': 'read_status',
                'args': {},
            },
            extra_rpc_args={},
            mock_api_response={})

    def test_check_position_false(self):
        '''Test check_position command: not at position'''
        def exec_command():
            self.fb.state.last_messages['status'] = [{
                'topic': '',
                'content': {
                    'location_data': {'position': {'x': 1, 'y': 2, 'z': 3}},
                },
            }]
            at_position = self.fb.check_position({'x': 0, 'y': 0, 'z': 0}, 2)
            self.assertFalse(at_position)
        self.send_command_test_helper(
            exec_command,
            expected_command={
                'kind': 'read_status',
                'args': {},
            },
            extra_rpc_args={},
            mock_api_response={})

    def test_check_position_no_status(self):
        '''Test check_position command: no status'''
        def exec_command():
            self.fb.state.last_messages['status'] = []
            at_position = self.fb.check_position({'x': 0, 'y': 0, 'z': 0}, 2)
            self.assertFalse(at_position)
        self.send_command_test_helper(
            exec_command,
            expected_command={
                'kind': 'read_status',
                'args': {},
            },
            extra_rpc_args={},
            mock_api_response={})

    def test_mount_tool(self):
        '''Test mount_tool command'''
        def exec_command():
            self.fb.mount_tool('Weeder')
        self.send_command_test_helper(
            exec_command,
            expected_command={
                'kind': 'lua',
                'args': {'lua': 'mount_tool("Weeder")'},
            },
            extra_rpc_args={},
            mock_api_response={})

    def test_dismount_tool(self):
        '''Test dismount_tool command'''
        def exec_command():
            self.fb.dismount_tool()
        self.send_command_test_helper(
            exec_command,
            expected_command={
                'kind': 'lua',
                'args': {'lua': 'dismount_tool()'},
            },
            extra_rpc_args={},
            mock_api_response={})

    def test_water(self):
        '''Test water command'''
        def exec_command():
            self.fb.water(123)
        self.send_command_test_helper(
            exec_command,
            expected_command={
                'kind': 'lua',
                'args': {'lua': '''plant = api({
                method = "GET",
                url = "/api/points/123"
            })
            water(plant)'''},
            },
            extra_rpc_args={},
            mock_api_response={})

    def test_dispense(self):
        '''Test dispense command'''
        def exec_command():
            self.fb.dispense(100)
        self.send_command_test_helper(
            exec_command,
            expected_command={
                'kind': 'lua',
                'args': {
                    'lua': 'dispense(100)',
                },
            },
            extra_rpc_args={},
            mock_api_response={})

    def test_dispense_all_args(self):
        '''Test dispense command with all args'''
        def exec_command():
            self.fb.dispense(100, 'Nutrient Sprayer', 4)
        self.send_command_test_helper(
            exec_command,
            expected_command={
                'kind': 'lua',
                'args': {
                    'lua': 'dispense(100, {tool_name = "Nutrient Sprayer", pin = 4})',
                },
            },
            extra_rpc_args={},
            mock_api_response={})

    def test_dispense_only_pin(self):
        '''Test dispense command'''
        def exec_command():
            self.fb.dispense(100, pin=4)
        self.send_command_test_helper(
            exec_command,
            expected_command={
                'kind': 'lua',
                'args': {
                    'lua': 'dispense(100, {pin = 4})',
                },
            },
            extra_rpc_args={},
            mock_api_response={})

    def test_dispense_only_tool_name(self):
        '''Test dispense command'''
        def exec_command():
            self.fb.dispense(100, "Nutrient Sprayer")
        self.send_command_test_helper(
            exec_command,
            expected_command={
                'kind': 'lua',
                'args': {
                    'lua': 'dispense(100, {tool_name = "Nutrient Sprayer"})',
                },
            },
            extra_rpc_args={},
            mock_api_response={})

    @patch('requests.request')
    def helper_get_seed_tray_cell(self, *args, **kwargs):
        '''Test helper for get_seed_tray_cell command'''
        mock_request = args[0]
        tray_data = kwargs['tray_data']
        cell = kwargs['cell']
        expected_xyz = kwargs['expected_xyz']
        self.fb.clear_cache()
        mock_response = Mock()
        mock_api_response = [
            {
                'id': 123,
                'name': 'Seed Tray',
                'pointer_type': '',  # not an actual data field
            },
            {
                'pointer_type': 'ToolSlot',
                'pullout_direction': 1,
                'x': 0,
                'y': 0,
                'z': 0,
                'tool_id': 123,
                'name': '',  # not an actual data field
                **tray_data,
            },
        ]
        mock_response.json.return_value = mock_api_response
        mock_response.status_code = 200
        mock_response.text = 'text'
        mock_request.return_value = mock_response
        cell = self.fb.get_seed_tray_cell('Seed Tray', cell)
        mock_request.assert_has_calls([
            call(
                method='GET',
                url='https://my.farm.bot/api/tools',
                **REQUEST_KWARGS,
            ),
            call().json(),
            call(
                method='GET',
                url='https://my.farm.bot/api/points',
                **REQUEST_KWARGS,
            ),
            call().json(),
        ])
        self.assertEqual(cell, expected_xyz, kwargs)

    def test_get_seed_tray_cell(self):
        '''Test get_seed_tray_cell'''
        test_cases = [
            {
                'tray_data': {'pullout_direction': 1},
                'cell': 'a1',
                'expected_xyz': {'x': 1.25, 'y': -18.75, 'z': 0},
            },
            {
                'tray_data': {'pullout_direction': 1},
                'cell': 'b2',
                'expected_xyz': {'x': -11.25, 'y': -6.25, 'z': 0},
            },
            {
                'tray_data': {'pullout_direction': 1},
                'cell': 'd4',
                'expected_xyz': {'x': -36.25, 'y': 18.75, 'z': 0},
            },
            {
                'tray_data': {'pullout_direction': 2},
                'cell': 'a1',
                'expected_xyz': {'x': -36.25, 'y': 18.75, 'z': 0},
            },
            {
                'tray_data': {'pullout_direction': 2},
                'cell': 'b2',
                'expected_xyz': {'x': -23.75, 'y': 6.25, 'z': 0},
            },
            {
                'tray_data': {'pullout_direction': 2},
                'cell': 'd4',
                'expected_xyz': {'x': 1.25, 'y': -18.75, 'z': 0},
            },
            {
                'tray_data': {'pullout_direction': 2, 'x': 100, 'y': 200, 'z': -100},
                'cell': 'd4',
                'expected_xyz': {'x': 101.25, 'y': 181.25, 'z': -100},
            },
        ]
        for test_case in test_cases:
            self.helper_get_seed_tray_cell(**test_case)

    @patch('requests.request')
    def helper_get_seed_tray_cell_error(self, *args, **kwargs):
        '''Test helper for get_seed_tray_cell command errors'''
        mock_request = args[0]
        tray_data = kwargs['tray_data']
        cell = kwargs['cell']
        error = kwargs['error']
        mock_response = Mock()
        mock_api_response = [
            {
                'id': 123,
                'name': 'Seed Tray',
                'pointer_type': '',  # not an actual data field
            },
            {
                'pointer_type': 'ToolSlot',
                'pullout_direction': 1,
                'x': 0,
                'y': 0,
                'z': 0,
                'tool_id': 123,
                'name': '',  # not an actual data field
                **tray_data,
            },
        ]
        mock_response.json.return_value = mock_api_response
        mock_response.status_code = 200
        mock_response.text = 'text'
        mock_request.return_value = mock_response
        with self.assertRaises(ValueError) as cm:
            self.fb.get_seed_tray_cell('Seed Tray', cell)
        self.assertEqual(cm.exception.args[0], error)
        mock_request.assert_has_calls([
            call(
                method='GET',
                url='https://my.farm.bot/api/tools',
                **REQUEST_KWARGS,
            ),
            call().json(),
            call(
                method='GET',
                url='https://my.farm.bot/api/points',
                **REQUEST_KWARGS,
            ),
            call().json(),
        ])

    def test_get_seed_tray_cell_invalid_cell_name(self):
        '''Test get_seed_tray_cell: invalid cell name'''
        self.helper_get_seed_tray_cell_error(
            tray_data={},
            cell='e4',
            error='Seed Tray Cell must be one of **A1** through **D4**',
        )

    def test_get_seed_tray_cell_invalid_pullout_direction(self):
        '''Test get_seed_tray_cell: invalid pullout direction'''
        self.helper_get_seed_tray_cell_error(
            tray_data={'pullout_direction': 0},
            cell='d4',
            error='Seed Tray **SLOT DIRECTION** must be `Positive X` or `Negative X`',
        )

    @patch('requests.request')
    def test_get_seed_tray_cell_no_tray(self, mock_request):
        '''Test get_seed_tray_cell: no seed tray'''
        mock_response = Mock()
        mock_api_response = []
        mock_response.json.return_value = mock_api_response
        mock_response.status_code = 200
        mock_response.text = 'text'
        mock_request.return_value = mock_response
        result = self.fb.get_seed_tray_cell('Seed Tray', 'a1')
        mock_request.assert_has_calls([
            call(
                method='GET',
                url='https://my.farm.bot/api/tools',
                **REQUEST_KWARGS,
            ),
            call().json(),
        ])
        self.assertIsNone(result)

    @patch('requests.request')
    def test_get_seed_tray_cell_not_mounted(self, mock_request):
        '''Test get_seed_tray_cell: seed tray not mounted'''
        mock_response = Mock()
        mock_api_response = [{
            'id': 123,
            'name': 'Seed Tray',
            'pointer_type': '',  # not an actual data field,
        }]
        mock_response.json.return_value = mock_api_response
        mock_response.status_code = 200
        mock_response.text = 'text'
        mock_request.return_value = mock_response
        result = self.fb.get_seed_tray_cell('Seed Tray', 'a1')
        mock_request.assert_has_calls([
            call(
                method='GET',
                url='https://my.farm.bot/api/tools',
                **REQUEST_KWARGS,
            ),
            call().json(),
        ])
        self.assertIsNone(result)

    def test_get_job_one(self):
        '''Test get_job command: get one job'''
        def exec_command():
            self.fb.state.last_messages['status'] = [{
                'topic': '',
                'content': {
                    'jobs': {
                        'job name': {'status': 'working'},
                    },
                },
            }]
            job = self.fb.get_job('job name')
            self.assertEqual(job, {'status': 'working'})
        self.send_command_test_helper(
            exec_command,
            expected_command={
                'kind': 'read_status',
                'args': {},
            },
            extra_rpc_args={},
            mock_api_response={})

    def test_get_job_all(self):
        '''Test get_job command: get all jobs'''
        def exec_command():
            self.fb.state.last_messages['status'] = [{
                'topic': '',
                'content': {
                    'jobs': {
                        'job name': {'status': 'working'},
                    },
                },
            }]
            jobs = self.fb.get_job()
            self.assertEqual(jobs, {'job name': {'status': 'working'}})
        self.send_command_test_helper(
            exec_command,
            expected_command={
                'kind': 'read_status',
                'args': {},
            },
            extra_rpc_args={},
            mock_api_response={})

    def test_get_job_no_status(self):
        '''Test get_job command: no status'''
        def exec_command():
            self.fb.state.last_messages['status'] = []
            job = self.fb.get_job('job name')
            self.assertIsNone(job)
        self.send_command_test_helper(
            exec_command,
            expected_command={
                'kind': 'read_status',
                'args': {},
            },
            extra_rpc_args={},
            mock_api_response={})

    def test_set_job(self):
        '''Test set_job command'''
        def exec_command():
            self.fb.set_job('job name', 'working', 50)
        self.send_command_test_helper(
            exec_command,
            expected_command={
                'kind': 'lua',
                'args': {'lua': '''local job_name = "job name"
            set_job(job_name)

            -- Update the job's status and percent:
            set_job(job_name, {
            status = "working",
            percent = 50
            })'''},
            },
            extra_rpc_args={},
            mock_api_response={})

    def test_complete_job(self):
        '''Test complete_job command'''
        def exec_command():
            self.fb.complete_job('job name')
        self.send_command_test_helper(
            exec_command,
            expected_command={
                'kind': 'lua',
                'args': {'lua': 'complete_job("job name")'},
            },
            extra_rpc_args={},
            mock_api_response={})

    def test_lua(self):
        '''Test lua command'''
        def exec_command():
            self.fb.lua('return true')
        self.send_command_test_helper(
            exec_command,
            expected_command={
                'kind': 'lua',
                'args': {'lua': 'return true'},
            },
            extra_rpc_args={},
            mock_api_response={})

    def test_if_statement(self):
        '''Test if_statement command'''
        def exec_command():
            self.fb.if_statement('pin10', 'is', 0)
        self.send_command_test_helper(
            exec_command,
            expected_command={
                'kind': '_if',
                'args': {
                    'lhs': 'pin10',
                    'op': 'is',
                    'rhs': 0,
                    '_then': {'kind': 'nothing', 'args': {}},
                    '_else': {'kind': 'nothing', 'args': {}},
                }
            },
            extra_rpc_args={},
            mock_api_response=[])

    def test_if_statement_with_named_pin(self):
        '''Test if_statement command with named pin'''
        def exec_command():
            self.fb.if_statement(
                'Lights', 'is', 0,
                named_pin_type='Peripheral')
        self.send_command_test_helper(
            exec_command,
            expected_command={
                'kind': '_if',
                'args': {
                    'lhs': {
                        'kind': 'named_pin',
                        'args': {'pin_type': 'Peripheral', 'pin_id': 123},
                    },
                    'op': 'is',
                    'rhs': 0,
                    '_then': {'kind': 'nothing', 'args': {}},
                    '_else': {'kind': 'nothing', 'args': {}},
                }
            },
            extra_rpc_args={},
            mock_api_response=[{'id': 123, 'label': 'Lights', 'mode': 0}])

    def test_if_statement_with_named_pin_not_found(self):
        '''Test if_statement command: named pin not found'''
        def exec_command():
            self.fb.if_statement(
                'Lights', 'is', 0,
                named_pin_type='Peripheral')
        self.send_command_test_helper(
            exec_command,
            expected_command=None,
            extra_rpc_args={},
            mock_api_response=[{'label': 'Pump'}])
        self.assertEqual(
            self.fb.state.error,
            "ERROR: 'Lights' not in peripherals: ['Pump'].")

    def test_if_statement_with_sequences(self):
        '''Test if_statement command with sequences'''
        def exec_command():
            self.fb.if_statement(
                'pin10', '<', 0,
                'Watering Sequence',
                'Drying Sequence')
        self.send_command_test_helper(
            exec_command,
            expected_command={
                'kind': '_if',
                'args': {
                    'lhs': 'pin10',
                    'op': '<',
                    'rhs': 0,
                    '_then': {'kind': 'execute', 'args': {'sequence_id': 123}},
                    '_else': {'kind': 'execute', 'args': {'sequence_id': 456}},
                }
            },
            extra_rpc_args={},
            mock_api_response=[
                {'id': 123, 'name': 'Watering Sequence'},
                {'id': 456, 'name': 'Drying Sequence'},
            ])

    def test_if_statement_with_sequence_not_found(self):
        '''Test if_statement command: sequence not found'''
        def exec_command():
            self.fb.if_statement(
                'pin10', '<', 0,
                'Watering Sequence',
                'Drying Sequence')
        self.send_command_test_helper(
            exec_command,
            expected_command=None,
            extra_rpc_args={},
            mock_api_response=[])
        self.assertEqual(
            self.fb.state.error,
            "ERROR: 'Watering Sequence' not in sequences: [].")

    def test_if_statement_invalid_operator(self):
        '''Test if_statement command: invalid operator'''
        def exec_command():
            with self.assertRaises(ValueError) as cm:
                self.fb.if_statement('pin10', 'nope', 0)
            self.assertEqual(
                cm.exception.args[0],
                "Invalid operator: nope not in ['<', '>', 'is', 'not', 'is_undefined']")
        self.send_command_test_helper(
            exec_command,
            expected_command=None,
            extra_rpc_args={},
            mock_api_response=[])

    def test_if_statement_invalid_variable(self):
        '''Test if_statement command: invalid variable'''
        variables = ["x", "y", "z", *[f"pin{str(i)}" for i in range(70)]]

        def exec_command():
            with self.assertRaises(ValueError) as cm:
                self.fb.if_statement('nope', '<', 0)
            self.assertEqual(
                cm.exception.args[0],
                f"Invalid variable: nope not in {variables}")
        self.send_command_test_helper(
            exec_command,
            expected_command=None,
            extra_rpc_args={},
            mock_api_response=[])

    def test_if_statement_invalid_named_pin_type(self):
        '''Test if_statement command: invalid named pin type'''
        def exec_command():
            with self.assertRaises(ValueError) as cm:
                self.fb.if_statement('pin10', '<', 0, named_pin_type='nope')
            self.assertEqual(
                cm.exception.args[0],
                "Invalid named_pin_type: nope not in ['Peripheral', 'Sensor']")
        self.send_command_test_helper(
            exec_command,
            expected_command=None,
            extra_rpc_args={},
            mock_api_response=[])

    def test_rpc_error(self):
        '''Test rpc error handling'''
        def exec_command():
            self.fb.wait(100)
            self.assertEqual(
                self.fb.state.error,
                'RPC error response received.')
        self.send_command_test_helper(
            exec_command,
            error=True,
            expected_command={
                'kind': 'wait',
                'args': {'milliseconds': 100}},
            extra_rpc_args={},
            mock_api_response=[])

    def test_rpc_response_timeout(self):
        '''Test rpc response timeout handling'''
        def exec_command():
            self.fb.state.last_messages['from_device'] = []
            self.fb.wait(100)
            self.assertEqual(
                self.fb.state.error,
                'Timed out waiting for RPC response.')
        self.send_command_test_helper(
            exec_command,
            expected_command={
                'kind': 'wait',
                'args': {'milliseconds': 100}},
            extra_rpc_args={},
            mock_api_response=[])

    def test_set_verbosity(self):
        '''Test set_verbosity.'''
        self.assertEqual(self.fb.state.verbosity, 0)
        self.fb.set_verbosity(1)
        self.assertEqual(self.fb.state.verbosity, 1)

    def test_set_timeout(self):
        '''Test set_timeout.'''
        self.assertEqual(self.fb.state.timeout['listen'], 0)
        self.fb.set_timeout(15)
        self.assertEqual(self.fb.state.timeout['listen'], 15)

    @staticmethod
    def helper_get_print_strings(mock_print):
        '''Test helper to get print call strings.'''
        return [string[1][0] for string in mock_print.mock_calls if len(string[1]) > 0]

    @patch('builtins.print')
    def test_print_status(self, mock_print):
        '''Test print_status.'''
        self.fb.set_verbosity(0)
        self.fb.state.print_status(description="testing")
        mock_print.assert_not_called()
        self.fb.set_verbosity(1)
        self.fb.state.print_status(description="testing")
        call_strings = self.helper_get_print_strings(mock_print)
        self.assertIn('testing', call_strings)
        mock_print.reset_mock()
        self.fb.set_verbosity(2)
        self.fb.state.print_status(endpoint_json=["testing"])
        call_strings = self.helper_get_print_strings(mock_print)
        call_strings = [s.split('(')[0].strip('`') for s in call_strings]
        self.assertIn('[\n    "testing"\n]', call_strings)
        self.assertIn('test_print_status', call_strings)



"""
ToolControls class.
"""

# └── functions/tools.py
#     ├── [BROKER] verify_tool()
#     ├── [BROKER] mount_tool()
#     ├── [BROKER] dismount_tool()
#     ├── [BROKER] water()
#     └── [BROKER] dispense()

from .broker import BrokerConnect
from .resources import Resources


class ToolControls():
    """Tool controls class."""

    def __init__(self, state):
        self.broker = BrokerConnect(state)
        self.resource = Resources(state)
        self.state = state

    # TODO: verify_tool()

    def mount_tool(self, tool_name):
        """Mounts the given tool and pulls it out of assigned slot."""
        self.state.print_status(description=f"Mounting {tool_name} tool.")

        lua_code = f"mount_tool(\"{tool_name}\")"

        self.resource.lua(lua_code)

    def dismount_tool(self):
        """Dismounts the currently mounted tool into assigned slot."""
        self.state.print_status(description="Dismounting tool.")

        lua_code = "dismount_tool()"

        self.resource.lua(lua_code)

    @staticmethod
    def water_and_dispense_end_string(tool_name=None, pin=None):
        """Prepares water or dispense end string."""
        string = ""
        if tool_name is not None:
            string += ", {"
            string += f"tool_name = \"{tool_name}\""
        if pin is not None:
            string += ", "
            if tool_name is None:
                string += "{"
            string += f"pin = {pin}"
        if tool_name is not None or pin is not None:
            string += "})"
        else:
            string += ")"
        return string

    def water(self, plant_id, tool_name=None, pin=None):
        """Moves to and waters plant based on age and assigned watering curve."""
        self.state.print_status(description=f"Watering plant {plant_id}...")

        lua_code = f"""
            plant = api({{
                method = "GET",
                url = "/api/points/{plant_id}"
            }})
            water(plant
        """
        lua_code = lua_code.strip()
        lua_code += self.water_and_dispense_end_string(tool_name, pin)

        self.resource.lua(lua_code)

    def dispense(self, milliliters, tool_name=None, pin=None):
        """Dispenses user-defined amount of liquid in milliliters."""
        self.state.print_status(
            description=f"Dispensing {milliliters} from tool {tool_name}...")

        lua_code = f"dispense({milliliters}"
        lua_code += self.water_and_dispense_end_string(tool_name, pin)

        self.resource.lua(lua_code)



"""
Information class.
"""

# └── functions/information
#     ├── [API] api_get()
#     ├── [API] api_patch()
#     ├── [API] api_post()
#     ├── [API] api_delete()
#     ├── [API] safe_z()
#     ├── [API] garden_size()
#     ├── [API] curve()
#     ├── [BROKER] measure_soil_height()
#     ├── [BROKER] read_status()
#     ├── [BROKER] read_pin()
#     └── [BROKER] read_sensor()

from .broker import BrokerConnect
from .api import ApiConnect


class Information():
    """Information class."""

    def __init__(self, state):
        self.broker = BrokerConnect(state)
        self.api = ApiConnect(state)
        self.state = state

    def api_get(self, endpoint, database_id=None, payload=None, data_print=True):
        """Get information about a specific endpoint."""
        self.state.print_status(
            description=f"Retrieving {endpoint} information.")

        endpoint_data = self.api.request(
            method="GET",
            endpoint=endpoint,
            database_id=database_id,
            payload=payload)

        if data_print:
            self.state.print_status(
                update_only=True,
                endpoint_json=endpoint_data)
        else:
            self.state.print_status(
                update_only=True,
                description=f"Fetched {len(endpoint_data)} items.")

        return endpoint_data

    def api_patch(self, endpoint, payload, database_id=None):
        """Change information contained within an endpoint."""
        self.state.print_status(description=f"Editing {endpoint}.")

        result = self.api.request(
            method="PATCH",
            endpoint=endpoint,
            database_id=database_id,
            payload=payload)

        self.state.print_status(update_only=True, endpoint_json=result)

        return result

    def api_post(self, endpoint, payload=None):
        """Create new information contained within an endpoint."""
        self.state.print_status(description=f"Adding new data to {endpoint}.")

        result = self.api.request(
            method="POST",
            endpoint=endpoint,
            database_id=None,
            payload=payload)

        self.state.print_status(update_only=True, endpoint_json=result)

        return result

    def api_delete(self, endpoint, database_id=None, payload=None):
        """Delete information contained within an endpoint."""
        self.state.print_status(
            description=f"Deleting {endpoint} with id={database_id}.")

        result = self.api.request(
            method="DELETE",
            endpoint=endpoint,
            database_id=database_id,
            payload=payload)

        self.state.print_status(update_only=True, endpoint_json=result)

        return result

    def safe_z(self):
        """Returns the highest safe point along the z-axis."""
        self.state.print_status(description="Retrieving safe z value...")

        config_data = self.api_get('fbos_config')
        z_value = config_data["safe_height"]

        self.state.print_status(
            description=f"Safe z={z_value}", update_only=True)
        return z_value

    def garden_size(self):
        """Return size of garden bed."""
        self.state.print_status(description="Retrieving garden size...")

        json_data = self.api_get('firmware_config')

        x_steps = json_data['movement_axis_nr_steps_x']
        x_mm = json_data['movement_step_per_mm_x']

        y_steps = json_data['movement_axis_nr_steps_y']
        y_mm = json_data['movement_step_per_mm_y']

        z_steps = json_data['movement_axis_nr_steps_z']
        z_mm = json_data['movement_step_per_mm_z']

        garden_size = {
            "x": x_steps / x_mm,
            "y": y_steps / y_mm,
            "z": z_steps / z_mm,
        }

        self.state.print_status(endpoint_json=garden_size, update_only=True)
        return garden_size

    def get_curve(self, curve_id):
        """Retrieve curve data from the API and return a curve object with extras."""
        self.state.print_status(description="Preparing curve information...")

        api_curve_data = self.api_get("curves", curve_id)
        if isinstance(api_curve_data, str):
            return None
        return Curve(api_curve_data)

    def measure_soil_height(self):
        """Use the camera to measure the soil height at the current location."""
        self.state.print_status(description="Measuring soil height...")

        measure_soil_height_message = {
            "kind": "execute_script",
            "args": {
                "label": "Measure Soil Height"
            }
        }

        self.broker.publish(measure_soil_height_message)

    def read_status(self, path=None):
        """Returns the FarmBot status tree."""
        path_str = "" if path is None else f" of {path}"
        self.state.print_status(description=f"Reading status{path_str}...")
        status_message = {
            "kind": "read_status",
            "args": {}
        }
        self.broker.publish(status_message)

        status_trees = self.state.last_messages.get("status", [])
        status_tree = None
        if len(status_trees) > 0:
            status_tree = status_trees[-1]["content"]

        if path is not None:
            for key in path.split("."):
                status_tree = status_tree[key]

        self.state.print_status(update_only=True, endpoint_json=status_tree)
        return status_tree

    @staticmethod
    def convert_mode_to_number(mode):
        """Converts mode string to mode number."""
        modes = ["digital", "analog"]
        if str(mode).lower() not in modes:
            raise ValueError(f"Invalid mode: {mode} not in {modes}")
        return 0 if mode.lower() == "digital" else 1

    def read_pin(self, pin_number, mode="digital"):
        """Reads the given pin by number."""
        pin_mode = self.convert_mode_to_number(mode)
        self.state.print_status(
            description=f"Reading pin {pin_number} ({mode})...")
        read_pin_message = {
            "kind": "read_pin",
            "args": {
                "pin_number": pin_number,
                "label": "---",
                "pin_mode": pin_mode,
            }
        }
        self.broker.publish(read_pin_message)

    def read_sensor(self, sensor_name):
        """Reads the given sensor."""
        self.state.print_status(description=f"Reading {sensor_name} sensor...")
        sensor = self.get_resource_by_name("sensors", sensor_name)
        if sensor is None:
            return
        sensor_id = sensor["id"]
        mode = sensor["mode"]

        sensor_message = {
            "kind": "read_pin",
            "args": {
                "pin_mode": mode,
                "label": "---",
                "pin_number": {
                    "kind": "named_pin",
                    "args": {
                        "pin_type": "Sensor",
                        "pin_id": sensor_id,
                    }
                }
            }
        }

        self.broker.publish(sensor_message)

    def get_resource_by_name(self, endpoint, resource_name, name_key="label", query=None):
        """Find a resource by name."""
        self.state.print_status(
            description=f"Searching for {resource_name} in {endpoint}.")
        resources = self.state.fetch_cache(endpoint)
        if resources is None:
            resources = self.api_get(endpoint, data_print=False)
        else:
            self.state.print_status(
                description=f"Using {len(resources)} cached items.")
        if query is not None:
            for key, value in query.items():
                resources = [res for res in resources if res[key] == value]
        names = [resource[name_key] for resource in resources]
        if resource_name not in names:
            error = f"ERROR: '{resource_name}' not in {endpoint}: {names}."
            self.state.print_status(description=error, update_only=True)
            self.state.error = error
            self.state.clear_cache(endpoint)
            return None

        self.state.save_cache(endpoint, resources)
        resource = [p for p in resources if p[name_key] == resource_name][0]
        return resource


class Curve:
    """Curve data object for the get_curve() function to return."""

    def __init__(self, curve_data):
        self.curve_data = curve_data
        self.name = curve_data["name"]
        self.type = curve_data["type"]
        self.unit = "mL" if self.type == "water" else "mm"

    def __getitem__(self, key):
        """Allow dictionary-style access to attributes."""
        return getattr(self, key)

    def day(self, day):
        """Calculate the value for a specific day based on the curve data."""
        day = int(day)
        data = self.curve_data["data"]
        data = {int(key): val for key, val in data.items()}
        value = data.get(day)
        if value is not None:
            return value

        sorted_day_keys = sorted(data.keys())
        prev_day = None
        next_day = None
        for day_key in sorted_day_keys:
            if day_key < day:
                prev_day = day_key
            elif day_key > day and next_day is None:
                next_day = day_key
                break

        if prev_day is None:
            return data[sorted_day_keys[0]]

        if next_day is None:
            return data[sorted_day_keys[-1]]

        exact_value = (data[prev_day] * (next_day - day) +
                       data[next_day] * (day - prev_day)
                       ) / (next_day - prev_day)
        return round(exact_value, 2)



"""
BrokerConnect class.
"""

# └── functions/broker.py
#     ├── [BROKER] connect()
#     ├── [BROKER] disconnect()
#     ├── [BROKER] publish()
#     ├── [BROKER] start_listen()
#     ├── [BROKER] stop_listen()
#     └── [BROKER] listen()

import time
import math
import json
import uuid
import functools
from datetime import datetime
import paho.mqtt.client as mqtt


class BrokerConnect():
    """Broker connection class."""

    def __init__(self, state):
        self.state = state
        self.client = None

    def connect(self):
        """Establish persistent connection to send messages via message broker."""

        self.state.check_token()

        self.client = mqtt.Client()
        self.client.username_pw_set(
            username=self.state.token["token"]["unencoded"]["bot"],
            password=self.state.token["token"]["encoded"]
        )

        self.client.connect(
            self.state.token["token"]["unencoded"]["mqtt"],
            port=1883,
            keepalive=60
        )

        self.client.loop_start()

        self.state.print_status(description="Connected to message broker.")

    def disconnect(self):
        """Disconnect from the message broker."""

        if self.client is not None:
            self.client.loop_stop()
            self.client.disconnect()
            description = "Disconnected from message broker."
            self.state.print_status(description=description)

    def wrap_message(self, message, priority=None):
        """Wrap message in CeleryScript format."""
        rpc = {
            "kind": "rpc_request",
            "args": {
                "label": "",
            },
            "body": [message],
        }

        if priority is not None:
            rpc["args"]["priority"] = priority

        return rpc

    def publish(self, message):
        """Publish messages containing CeleryScript via the message broker."""

        if self.client is None:
            self.connect()

        rpc = message
        if rpc["kind"] != "rpc_request":
            rpc = self.wrap_message(rpc)

        if rpc["args"]["label"] == "":
            if self.state.test_env:
                rpc["args"]["label"] = "test"
            else:
                rpc["args"]["label"] = uuid.uuid4().hex

        self.state.print_status(description="Publishing to 'from_clients'")
        self.state.print_status(endpoint_json=rpc, update_only=True)
        if self.state.dry_run:
            self.state.print_status(
                description="Sending disabled, message not sent.",
                update_only=True)
        else:
            self.listen("from_device", publish_payload=rpc)

        response = self.state.last_messages.get("from_device", [])
        if len(response) > 0:
            if response[-1]["content"]["kind"] == "rpc_ok":
                self.state.print_status(
                    description="Success response received.",
                    update_only=True)
                self.state.error = None
            else:
                self.state.print_status(
                    description="Error response received.",
                    update_only=True)
                self.state.error = "RPC error response received."

        self.state.last_published = rpc

    def start_listen(self, channel="#", message_options=None):
        """Establish persistent subscription to message broker channels."""
        options = message_options or {}
        path = (options.get("path", "") or "").split(".") or []
        path = [key for key in path if key != ""]
        diff_only = options.get("diff_only")
        filters = options.get("filters", {})

        if self.client is None:
            self.connect()

        def add_message(key, topic, content):
            """Add message to last_messages."""
            if key not in self.state.last_messages:
                self.state.last_messages[key] = []
            if topic is None:
                self.state.last_messages[key].append(content)
                return
            self.state.last_messages[key].append({
                "topic": topic,
                "content": content,
            })

        # Set on_message callback
        def on_message(_client, _userdata, msg):
            """on_message callback"""
            channel_key = msg.topic.split("/")[2]
            payload = json.loads(msg.payload)

            if not self.match({"topic": msg.topic, "content": payload}, filters):
                self.state.print_status(
                    description="x",
                    update_only=True,
                    end="")
                return

            if channel == "#":
                add_message(channel, msg.topic, payload)
            add_message(channel_key, msg.topic, payload)

            for key in path:
                payload = payload[key]
            path_channel = f"{channel_key}_excerpt"
            if len(path) > 0:
                add_message(path_channel, None, payload)

            if diff_only:
                key = path_channel if len(path) > 0 else channel_key
                last_messages = self.state.last_messages.get(key, [])
                if len(last_messages) > 1:
                    if len(path) > 0:
                        current = last_messages[-1]
                        previous = last_messages[-2]
                    else:
                        current = last_messages[-1]["content"]
                        previous = last_messages[-2]["content"]
                    diff, _is_different = difference(current, previous)
                    payload = diff
                    add_message(f"{channel_key}_diffs", None, payload)

            self.state.print_status(description="", update_only=True)
            description = "New message"
            if len(path) > 0:
                description += f" {'.'.join(path)}"
            if diff_only:
                description += " diff"
            description += f" from {msg.topic}"
            timestamp = datetime.now().strftime("%Y-%m-%d %H:%M:%S")
            description += f" ({timestamp})"
            self.state.print_status(
                endpoint_json=payload,
                description=description)

        self.client.on_message = on_message

        # Subscribe to channel
        device_id_str = self.state.token["token"]["unencoded"]["bot"]
        self.client.subscribe(f"bot/{device_id_str}/{channel}")
        description = f"Connected to message broker channel '{channel}'"
        if channel == "#":
            description = "Connected to all message broker channels"
        self.state.print_status(description=description)

        # Start listening
        self.client.loop_start()
        description = f"Now listening to message broker channel '{channel}'"
        if channel == "#":
            description = "Now listening to all message broker channels"
        self.state.print_status(description=description)

    def stop_listen(self):
        """End subscription to all message broker channels."""

        self.client.loop_stop()

        self.state.print_status(
            description="Stopped listening to all message broker channels.")

    def clear_last_messages(self, key):
        """Clear last messages from a channel."""
        if key == "#":
            self.state.last_messages = {"#": []}
            return
        self.state.last_messages[key] = []
        self.state.last_messages[f"{key}_excerpt"] = []
        self.state.last_messages[f"{key}_diffs"] = []

    def match(self, message, filters):
        """Check if message matches filters."""
        if filters.get("topic", '') not in message["topic"]:
            return False
        for path, value in filters.get("content", {}).items():
            content = message["content"]
            for key in path.split("."):
                content = content[key]
            if str(value) not in str(content):
                return False
        return True

    @staticmethod
    def stop_listen_upon_interrupt(func):
        """Decorator to stop listening upon KeyboardInterrupt."""
        @functools.wraps(func)
        def wrapper(self, *args, **kwargs):
            try:
                func(self, *args, **kwargs)
            except KeyboardInterrupt:
                self.stop_listen()
        return wrapper

    @stop_listen_upon_interrupt.__func__
    def listen(self,
               channel="#",
               duration=None,
               publish_payload=None,
               stop_count=1,
               message_options=None):
        """Listen to a message broker channel for the provided duration in seconds."""
        publish = publish_payload is not None
        message = (publish_payload or {}).get("body", [{}])[0]
        message_options = message_options or {}
        filters = message_options.get("filters", {})
        filters = {"topic": '', "content": {}, **filters}
        # Prepare duration option
        timeout_key = "listen"
        if message.get("kind") in ["move", "find_home", "calibrate"]:
            timeout_key = "movements"
        duration_seconds = duration or self.state.timeout[timeout_key]
        if message.get("kind") == "wait":
            duration_seconds += message["args"]["milliseconds"] / 1000
        if stop_count > 1:
            duration_seconds = math.inf
        # Prepare label matching
        if publish and publish_payload["args"]["label"] != "":
            # If a label is provided, verify the label matches
            filters["content"]["args.label"] = publish_payload["args"]["label"]
        if message.get("kind") == "read_status":
            # Getting the RPC response to read_status isn't as important as
            # returning the status as soon as possible, since the device
            # will still publish a status within a few seconds even if the
            # read_status command isn't received.
            channel = "status"
            filters = {"topic": 'status', "content": {}}

        # Print status message
        description = "Listening to message broker"
        if channel != "#":
            description += f" channel '{channel}'"
        if duration_seconds != math.inf:
            description += f" for {duration_seconds} seconds"
        filter_strs = []
        if filters["topic"]:
            filter_strs.append(f" for `{filters['topic']}` in topic")
        for k, v in filters["content"].items():
            filter_strs.append(f" for `{v}` in `{k}`")
        description += " and".join(filter_strs)
        if message_options.get("diff_only"):
            description += " for diffs"
        if message_options.get("path"):
            description += f" at path `{message_options['path']}`"
        plural = "s are" if stop_count > 1 else " is"
        description += f" until {stop_count} message{plural} received"
        description += "..."
        self.state.print_status(description=description)

        # Start listening
        start_time = datetime.now()
        self.start_listen(channel, {**message_options, "filters": filters})
        if not self.state.test_env:
            self.clear_last_messages(channel)
        if publish:
            time.sleep(0.1)  # wait for start_listen to be ready
            device_id_str = self.state.token["token"]["unencoded"]["bot"]
            publish_topic = f"bot/{device_id_str}/from_clients"
            self.client.publish(
                publish_topic,
                payload=json.dumps(publish_payload))
        self.state.print_status(update_only=True, description="", end="")
        while (datetime.now() - start_time).seconds < duration_seconds:
            self.state.print_status(update_only=True, description=".", end="")
            time.sleep(0.25)
            last_messages = self.state.last_messages.get(channel, [])
            if len(last_messages) > (stop_count - 1):
                seconds = (datetime.now() - start_time).seconds
                prefix = f"{stop_count} messages"
                if stop_count == 1:
                    prefix = "Message"
                description = f"{prefix} received after {seconds} seconds"
                self.state.print_status(
                    description=description,
                    update_only=True)
                break
        if len(self.state.last_messages.get(channel, [])) == 0:
            self.state.print_status(description="", update_only=True)
            secs = duration_seconds
            description = f"Did not receive message after {secs} seconds"
            self.state.print_status(
                description=description,
                update_only=True)
            self.state.error = "Timed out waiting for RPC response."
        else:
            self.state.error = None

        self.stop_listen()


def difference(next_state, prev_state):
    """Find the difference between two states."""
    is_different = False
    diff = {}

    if next_state != prev_state:
        is_different = True

    if not isinstance(next_state, dict) or not isinstance(prev_state, dict):
        return next_state, is_different

    for key, next_value in next_state.items():
        if key not in prev_state:
            diff[key] = next_value
            is_different = True
            continue
        prev_value = prev_state[key]
        if next_value != prev_value:
            if isinstance(next_value, dict) and isinstance(prev_value, dict):
                nested_diff, nested_is_different = difference(
                    next_value,
                    prev_value)
                if nested_is_different:
                    diff[key] = nested_diff
                    is_different = True
            else:
                diff[key] = next_value
                is_different = True

    for key in prev_state:
        if key not in next_state:
            is_different = True

    return diff, is_different



"""
Farmbot class.
"""

from .state import State
from .functions.api import ApiConnect
from .functions.basic_commands import BasicCommands
from .functions.broker import BrokerConnect
from .functions.camera import Camera
from .functions.information import Information
from .functions.jobs import JobHandling
from .functions.messages import MessageHandling
from .functions.movements import MovementControls
from .functions.peripherals import Peripherals
from .functions.resources import Resources
from .functions.tools import ToolControls

VERSION = "2.0.5"


class Farmbot():
    """Farmbot class."""
    __version__ = VERSION

    def __init__(self):
        self.state = State()

        # Initialize other components without the token initially
        self.api = ApiConnect(self.state)
        self.basic = BasicCommands(self.state)
        self.broker = BrokerConnect(self.state)
        self.camera = Camera(self.state)
        self.info = Information(self.state)
        self.jobs = JobHandling(self.state)
        self.messages = MessageHandling(self.state)
        self.movements = MovementControls(self.state)
        self.peripherals = Peripherals(self.state)
        self.resources = Resources(self.state)
        self.tools = ToolControls(self.state)

    def set_verbosity(self, value):
        """Set output verbosity level."""
        self.state.verbosity = value

    def set_timeout(self, duration, key="listen"):
        """Set timeout value in seconds."""
        if key == "all":
            for timeout_key in self.state.timeout:
                self.state.timeout[timeout_key] = duration
        else:
            self.state.timeout[key] = duration

    def set_token(self, token):
        """Set FarmBot authorization token."""
        self.state.token = token

    def clear_cache(self, endpoint=None):
        """Clear cached records."""
        self.state.clear_cache(endpoint)

    # api.py

    def get_token(self, email, password, server="https://my.farm.bot"):
        """Get FarmBot authorization token. Server is 'https://my.farm.bot' by default."""
        return self.api.get_token(email, password, server)

    # basic_commands.py

    def wait(self, duration):
        """Pauses execution for a certain number of milliseconds."""
        return self.basic.wait(duration)

    def e_stop(self):
        """Emergency locks (E-stops) the Farmduino microcontroller."""
        return self.basic.e_stop()

    def unlock(self):
        """Unlocks a locked (E-stopped) device."""
        return self.basic.unlock()

    def reboot(self):
        """Reboots the FarmBot OS and re-initializes the device."""
        return self.basic.reboot()

    def shutdown(self):
        """Shuts down the FarmBot OS and turns the device off."""
        return self.basic.shutdown()

    # broker.py

    def connect_broker(self):
        """Establish persistent connection to send messages via message broker."""
        return self.broker.connect()

    def disconnect_broker(self):
        """Disconnect from the message broker."""
        return self.broker.disconnect()

    def publish(self, message):
        """Publish message to the message broker."""
        return self.broker.publish(message)

    def listen(self,
               channel="#",
               duration=None,
               stop_count=1,
               message_options=None):
        """Listen to a message broker channel."""
        return self.broker.listen(
            channel=channel,
            duration=duration,
            stop_count=stop_count,
            message_options=message_options)

    def listen_for_status_changes(self,
                                  duration=None,
                                  stop_count=1,
                                  diff_only=True,
                                  path=None):
        """Listen for status changes."""
        return self.broker.listen(
            channel="status",
            duration=duration,
            stop_count=stop_count,
            message_options={
                "diff_only": diff_only,
                "path": path,
                "filters": {
                    "topic": "status",
                    "content": {},
                }},
        )

    # camera.py

    def calibrate_camera(self):
        """Performs camera calibration. This action will reset camera calibration settings."""
        return self.camera.calibrate_camera()

    def take_photo(self):
        """Takes photo using the device camera and uploads it to the web app."""
        return self.camera.take_photo()

    # information.py

    def api_get(self, endpoint, database_id=None, payload=None):
        """Get information about a specific endpoint."""
        return self.info.api_get(endpoint, database_id, payload)

    def api_patch(self, endpoint, payload, database_id=None):
        """Change information contained within an endpoint."""
        return self.info.api_patch(endpoint, payload, database_id)

    def api_post(self, endpoint, payload=None):
        """Create new information contained within an endpoint."""
        return self.info.api_post(endpoint, payload)

    def api_delete(self, endpoint, database_id=None, payload=None):
        """Delete information contained within an endpoint."""
        return self.info.api_delete(endpoint, database_id, payload)

    def safe_z(self):
        """Returns the highest safe point along the z-axis."""
        return self.info.safe_z()

    def garden_size(self):
        """Returns size of garden bed."""
        return self.info.garden_size()

    def get_curve(self, curve_id=None):
        """Returns the curve data."""
        return self.info.get_curve(curve_id)

    def measure_soil_height(self):
        """Use the camera to determine soil height at the current location."""
        return self.info.measure_soil_height()

    def read_status(self, path=None):
        """Returns the FarmBot status tree."""
        return self.info.read_status(path)

    def read_pin(self, pin_number, mode="digital"):
        """Reads the current value of the specified pin."""
        return self.info.read_pin(pin_number, mode)

    def read_sensor(self, sensor_name):
        """Reads the given sensor."""
        return self.info.read_sensor(sensor_name)

    # jobs.py

    def get_job(self, job_name=None):
        """Retrieves the status or details of the specified job."""
        return self.jobs.get_job(job_name)

    def set_job(self, job_name, status, percent):
        """Initiates or modifies job with given parameters."""
        return self.jobs.set_job(job_name, status, percent)

    def complete_job(self, job_name):
        """Marks job as completed and triggers any associated actions."""
        return self.jobs.complete_job(job_name)

    # messages.py

    def log(self, message_str, message_type="info", channels=None):
        """Sends new log message via the API."""
        return self.messages.log(message_str, message_type, channels)

    def send_message(self, message_str, message_type="info", channels=None):
        """Sends new log message via the message broker."""
        return self.messages.send_message(message_str, message_type, channels)

    def debug(self, message_str):
        """Sends debug message used for developer information or troubleshooting."""
        return self.messages.debug(message_str)

    def toast(self, message_str, message_type="info"):
        """Sends a message that pops up on the user interface briefly."""
        return self.messages.toast(message_str, message_type)

    # movements.py

    def move(self, x=None, y=None, z=None, safe_z=None, speed=None):
        """Moves to the specified (x, y, z) coordinate."""
        return self.movements.move(x, y, z, safe_z, speed)

    def set_home(self, axis="all"):
        """Sets the current position as the home position for a specific axis."""
        return self.movements.set_home(axis)

    def find_home(self, axis="all", speed=100):
        """Moves the device to the home position for a specified axis."""
        return self.movements.find_home(axis, speed)

    def find_axis_length(self, axis="all"):
        """Finds the length of a specified axis."""
        return self.movements.find_axis_length(axis)

    def get_xyz(self):
        """Returns the current (x, y, z) coordinates of the FarmBot."""
        return self.movements.get_xyz()

    def check_position(self, coordinate, tolerance):
        """Verifies position of the FarmBot within specified tolerance range."""
        return self.movements.check_position(coordinate, tolerance)

    # peripherals.py

    def control_servo(self, pin, angle):
        """Set servo angle between 0-180 degrees."""
        return self.peripherals.control_servo(pin, angle)

    def write_pin(self, pin_number, value, mode="digital"):
        """Writes a new value to the specified pin."""
        return self.peripherals.write_pin(pin_number, value, mode)

    def control_peripheral(self, peripheral_name, value, mode=None):
        """Set peripheral value and mode."""
        return self.peripherals.control_peripheral(peripheral_name, value, mode)

    def toggle_peripheral(self, peripheral_name):
        """Toggles the state of a specific peripheral between `on` and `off`."""
        return self.peripherals.toggle_peripheral(peripheral_name)

    def on(self, pin_number):
        """Turns specified pin number `on` (100%)."""
        return self.peripherals.on(pin_number)

    def off(self, pin_number):
        """Turns specified pin number `off` (0%)."""
        return self.peripherals.off(pin_number)

    # resources.py

    def sequence(self, sequence_name):
        """Executes a predefined sequence."""
        return self.resources.sequence(sequence_name)

    def get_seed_tray_cell(self, tray_name, tray_cell):
        """Identifies and returns the location of specified cell in the seed tray."""
        return self.resources.get_seed_tray_cell(tray_name, tray_cell)

    def detect_weeds(self):
        """Scans the garden to detect weeds."""
        return self.resources.detect_weeds()

    def lua(self, lua_code):
        """Executes custom Lua code snippets to perform complex tasks or automations."""
        return self.resources.lua(lua_code)

    def if_statement(self,
                     variable,
                     operator,
                     value,
                     then_sequence_name=None,
                     else_sequence_name=None,
                     named_pin_type=None):
        """Performs conditional check and executes actions based on the outcome."""
        return self.resources.if_statement(
            variable,
            operator,
            value,
            then_sequence_name,
            else_sequence_name,
            named_pin_type)

    def assertion(self, lua_code, assertion_type, recovery_sequence_name=None):
        """Evaluates an expression."""
        return self.resources.assertion(lua_code, assertion_type, recovery_sequence_name)

    # tools.py

    def mount_tool(self, tool_name):
        """Mounts the given tool and pulls it out of assigned slot."""
        return self.tools.mount_tool(tool_name)

    def dismount_tool(self):
        """Dismounts the currently mounted tool into assigned slot."""
        return self.tools.dismount_tool()

    def water(self, plant_id, tool_name=None, pin=None):
        """Moves to and waters plant based on age and assigned watering curve."""
        return self.tools.water(plant_id, tool_name, pin)

    def dispense(self, milliliters, tool_name=None, pin=None):
        """Dispenses user-defined amount of liquid in milliliters."""
        return self.tools.dispense(milliliters, tool_name, pin)



"""State management."""

import json
import inspect
from datetime import datetime


def get_call_stack_depth():
    """Return the depth of the current call stack."""
    depth = 0
    frame = inspect.currentframe()
    while frame:
        depth += 1
        frame = frame.f_back
    return depth


def get_function_call_info():
    """Return the name and given arguments of the function where this is called."""
    # back to print_status then back to the function that called print_status
    frame = inspect.currentframe().f_back.f_back
    func_name = frame.f_code.co_name
    args, _, _, values = inspect.getargvalues(frame)
    arg_strings = []
    for arg in args:
        if arg != "self":
            arg_strings.append(f"{arg}={repr(values[arg])}")
    arg_str = ", ".join(arg_strings)
    return f"{func_name}({arg_str})"


NO_TOKEN_ERROR = """
ERROR: You have no token, please call `get_token`
using your login credentials and the server you wish to connect to.
"""


class State():
    """State class."""

    NO_TOKEN_ERROR = NO_TOKEN_ERROR.replace("\n", "")

    def __init__(self):
        self.token = None
        self.error = None
        self.last_messages = {}
        self.last_published = {}
        self.verbosity = 1
        self.json_printing = True
        self.timeout = {
            "api": 15,
            "listen": 15,
            "movements": 120,
        }
        self.test_env = False
        self.ssl = True
        self.min_call_stack_depth = 100
        self.dry_run = False
        self.resource_cache = {}

    def print_status(self, endpoint_json=None, description=None, update_only=False, end="\n"):
        """Handle changes to output based on user-defined verbosity."""
        depth = get_call_stack_depth()
        if depth < self.min_call_stack_depth:
            self.min_call_stack_depth = depth
        top = depth == self.min_call_stack_depth
        no_end = end == "" and description != ""
        indent = "" if (top or no_end) else " " * 4

        if self.verbosity >= 2 and not update_only:
            if top:
                print()
            function = get_function_call_info()
            print(f"{indent}`{function}` called at {datetime.now()}")
        if self.verbosity >= 1:
            if self.verbosity == 1 and not update_only and top:
                print()
            if description is not None:
                print(indent + description, end=end, flush=end == "")
            if endpoint_json is not None and self.json_printing:
                json_str = json.dumps(endpoint_json, indent=4)
                indented_str = indent + json_str.replace("\n", "\n" + indent)
                print(indented_str)

    def check_token(self):
        """Check if a token is present."""
        if self.token is None:
            self.print_status(description=self.NO_TOKEN_ERROR)
            self.error = self.NO_TOKEN_ERROR
            raise ValueError(self.NO_TOKEN_ERROR)

    def save_cache(self, endpoint, records):
        """Cache records."""
        self.resource_cache[endpoint] = records

    def fetch_cache(self, endpoint):
        """Fetch cached records."""
        return self.resource_cache.get(endpoint)

    def clear_cache(self, endpoint=None):
        """Clear the cache."""
        if endpoint is not None and endpoint in self.resource_cache:
            del self.resource_cache[endpoint]
        else:
            self.resource_cache = {}



"""
Resources class.
"""

# └── functions/resources.py
#     ├── [BROKER] sort_points()
#     ├── [BROKER] sequence()
#     ├── [BROKER] get_seed_tray_cell()
#     ├── [BROKER] detect_weeds()
#     ├── [BROKER] lua()
#     ├── [BROKER] if_statement()
#     └── [BROKER] assertion()

from .broker import BrokerConnect
from .information import Information

ASSERTION_TYPES = ["abort", "recover", "abort_recover", "continue"]


def validate_assertion_type(assertion_type):
    """Validate assertion type."""
    if assertion_type not in ASSERTION_TYPES:
        msg = "Invalid assertion_type: "
        msg += f"{assertion_type} not in {ASSERTION_TYPES}"
        raise ValueError(msg)


OPERATORS = ["<", ">", "is", "not", "is_undefined"]
IF_STATEMENT_VARIABLE_STRINGS = [
    "x",
    "y",
    "z",
    *[f"pin{str(i)}" for i in range(70)]]
NAMED_PIN_TYPES = ["Peripheral", "Sensor"]


def validate_if_statement_args(named_pin_type, variable, operator):
    """Validate if statement arguments."""
    if operator not in OPERATORS:
        raise ValueError(f"Invalid operator: {operator} not in {OPERATORS}")
    if named_pin_type is None and variable not in IF_STATEMENT_VARIABLE_STRINGS:
        msg = "Invalid variable: "
        msg += f"{variable} not in {IF_STATEMENT_VARIABLE_STRINGS}"
        raise ValueError(msg)
    if named_pin_type is not None and named_pin_type not in NAMED_PIN_TYPES:
        msg = "Invalid named_pin_type: "
        msg += f"{named_pin_type} not in {NAMED_PIN_TYPES}"
        raise ValueError(msg)


class Resources():
    """Resources class."""

    def __init__(self, state):
        self.broker = BrokerConnect(state)
        self.info = Information(state)
        self.state = state

    # TODO: mark_as()

    # TODO: sort_points(points, method)

    def sequence(self, sequence_name):
        """Executes a predefined sequence."""
        self.state.print_status(
            description="Running {sequence_name} sequence.")

        sequence = self.info.get_resource_by_name(
            endpoint="sequences",
            resource_name=sequence_name,
            name_key="name")
        if sequence is None:
            return

        sequence_message = {
            "kind": "execute",
            "args": {
                "sequence_id": sequence["id"],
            }
        }

        self.broker.publish(sequence_message)

    def get_seed_tray_cell(self, tray_name, tray_cell):
        """Identifies and returns the location of specified cell in the seed tray."""
        self.state.print_status(
            description="Identifying seed tray cell location.")

        tray_tool = self.info.get_resource_by_name("tools", tray_name, "name")
        if tray_tool is None:
            return
        tray_data = self.info.get_resource_by_name(
            "points", tray_tool["id"], "tool_id", {"pointer_type": "ToolSlot"})
        if tray_data is None:
            self.state.print_status(
                description=f"{tray_name} must be mounted in a slot.",
                update_only=True)
            return

        cell = tray_cell.upper()

        seeder_needle_offset = 17.5
        cell_spacing = 12.5

        cells = {
            "A1": {"x": 0, "y": 0},
            "A2": {"x": 0, "y": 1},
            "A3": {"x": 0, "y": 2},
            "A4": {"x": 0, "y": 3},

            "B1": {"x": -1, "y": 0},
            "B2": {"x": -1, "y": 1},
            "B3": {"x": -1, "y": 2},
            "B4": {"x": -1, "y": 3},

            "C1": {"x": -2, "y": 0},
            "C2": {"x": -2, "y": 1},
            "C3": {"x": -2, "y": 2},
            "C4": {"x": -2, "y": 3},

            "D1": {"x": -3, "y": 0},
            "D2": {"x": -3, "y": 1},
            "D3": {"x": -3, "y": 2},
            "D4": {"x": -3, "y": 3}
        }

        if cell not in cells:
            msg = "Seed Tray Cell must be one of **A1** through **D4**"
            raise ValueError(msg)

        flip = 1
        if tray_data["pullout_direction"] == 1:
            flip = 1
        elif tray_data["pullout_direction"] == 2:
            flip = -1
        else:
            msg = "Seed Tray **SLOT DIRECTION** must be `Positive X` or `Negative X`"
            raise ValueError(msg)

        a1 = {
            "x": tray_data["x"] - seeder_needle_offset + (1.5 * cell_spacing * flip),
            "y": tray_data["y"] - (1.5 * cell_spacing * flip),
            "z": tray_data["z"]
        }

        offset = {
            "x": cell_spacing * cells[cell]["x"] * flip,
            "y": cell_spacing * cells[cell]["y"] * flip
        }

        cell_xyz = {
            "x": a1["x"] + offset["x"],
            "y": a1["y"] + offset["y"],
            "z": a1["z"],
        }

        self.state.print_status(
            description=f"Cell {tray_cell} is at {cell_xyz}.",
            update_only=True)
        return cell_xyz

    def detect_weeds(self):
        """Scans the garden to detect weeds."""
        self.state.print_status(description="Detecting weeds...")

        detect_weeds_message = {
            "kind": "execute_script",
            "args": {
                "label": "plant-detection"
            }
        }

        self.broker.publish(detect_weeds_message)

    def lua(self, lua_code):
        """Executes custom Lua code snippets to perform complex tasks or automations."""
        self.state.print_status(description="Running Lua code")

        lua_message = {
            "kind": "lua",
            "args": {
                "lua": lua_code.strip()
            }
        }

        self.broker.publish(lua_message)

    def if_statement(self,
                     variable,
                     operator,
                     value,
                     then_sequence_name=None,
                     else_sequence_name=None,
                     named_pin_type=None):
        """Performs conditional check and executes actions based on the outcome."""

        self.state.print_status(description="Executing if statement.")

        validate_if_statement_args(named_pin_type, variable, operator)
        if named_pin_type is not None:
            endpoint = named_pin_type.lower() + "s"
            resource = self.info.get_resource_by_name(endpoint, variable)
            if resource is None:
                return
            variable = {
                "kind": "named_pin",
                "args": {
                    "pin_type": named_pin_type,
                    "pin_id": resource["id"]
                }
            }

        if_statement_message = {
            "kind": "_if",
            "args": {
                "lhs": variable,
                "op": operator,
                "rhs": value,
                "_then": {"kind": "nothing", "args": {}},
                "_else": {"kind": "nothing", "args": {}},
            }
        }

        sequence_names = {
            "_then": then_sequence_name,
            "_else": else_sequence_name,
        }
        for key, sequence_name in sequence_names.items():
            if sequence_name is not None:
                sequence = self.info.get_resource_by_name(
                    endpoint="sequences",
                    resource_name=sequence_name,
                    name_key="name")
                if sequence is None:
                    return
                sequence_id = sequence["id"]
                if_statement_message["args"][key] = {
                    "kind": "execute",
                    "args": {"sequence_id": sequence_id},
                }

        self.broker.publish(if_statement_message)

    def assertion(self, lua_code, assertion_type, recovery_sequence_name=None):
        """Evaluates an expression."""
        self.state.print_status(description="Executing assertion.")

        validate_assertion_type(assertion_type)

        assertion_message = {
            "kind": "assertion",
            "args": {
                "assertion_type": assertion_type,
                "lua": lua_code,
                "_then": {"kind": "nothing", "args": {}},
            }
        }

        if recovery_sequence_name is not None:
            sequence = self.info.get_resource_by_name(
                endpoint="sequences",
                resource_name=recovery_sequence_name,
                name_key="name")
            if sequence is None:
                return
            recovery_sequence_id = sequence["id"]
            assertion_message["args"]["_then"] = {
                "kind": "execute",
                "args": {"sequence_id": recovery_sequence_id},
            }

        self.broker.publish(assertion_message)
